#include "cost_saturation.h"

#include "abstraction.h"
#include "additive_cartesian_heuristic.h"
#include "subtask_generators.h"
#include "transition_system.h"
#include "utils.h"

#include "../globals.h"

#include "../task_tools.h"
#include "../tasks/modified_operator_costs_task.h"
#include "../utils/countdown_timer.h"
#include "../utils/logging.h"
#include "../utils/memory.h"

#include "../option_parser.h"
#include "../plugin.h"
#include "../operator_cost.h"
#include "../utils/rng_options.h"

#include <algorithm>
#include <cassert>

using namespace std;

namespace cegar {
// TODO: Remove.
int hacked_num_landmark_abstractions = -1;

void reduce_costs(
    vector<int> &remaining_costs, const vector<int> &saturated_costs) {
    assert(remaining_costs.size() == saturated_costs.size());
    for (size_t i = 0; i < remaining_costs.size(); ++i) {
        int &remaining = remaining_costs[i];
        const int &saturated = saturated_costs[i];
        assert(saturated <= remaining);
        /* Since we ignore transitions from states s with h(s)=INF, all
           saturated costs (h(s)-h(s')) are finite or -INF. */
        assert(saturated != INF);
        if (remaining == INF) {
            // INF - x = INF for finite values x.
        } else if (saturated == -INF) {
            remaining = INF;
        } else {
            remaining -= saturated;
        }
        assert(remaining >= 0);
    }
}

CostSaturation::CostSaturation(
    CostPartitioningType cost_partitioning_type,
    const vector<shared_ptr<SubtaskGenerator>> &subtask_generators,
    int max_states,
    int max_non_looping_transitions,
    double max_time,
    bool use_general_costs,
    bool exclude_abstractions_with_zero_init_h,
    PickSplit pick_split,
    shared_ptr<utils::RandomNumberGenerator> rng,
    bool compress_cegar_states)
    : cost_partitioning_type(cost_partitioning_type),
      subtask_generators(subtask_generators),
      max_states(max_states),
      max_non_looping_transitions(max_non_looping_transitions),
      max_time(max_time),
      use_general_costs(use_general_costs),
      exclude_abstractions_with_zero_init_h(exclude_abstractions_with_zero_init_h),
      pick_split(pick_split),
      rng(rng),
      num_abstractions(0),
      num_states(0),
      num_non_looping_transitions(0),
      compress_cegar_states(compress_cegar_states) {
}

void CostSaturation::initialize(const shared_ptr<AbstractTask> &task) {
    utils::CountdownTimer timer(max_time);

    TaskProxy task_proxy(*task);

    verify_no_axioms(task_proxy);
    verify_no_conditional_effects(task_proxy);

    reset(task_proxy);

    function<bool()> should_abort =
        [&] () {
            return num_states >= max_states ||
                   num_non_looping_transitions >= max_non_looping_transitions ||
                   timer.is_expired() ||
                   utils::is_out_of_memory() ||
                   initial_state_is_dead_end();
        };

    for (shared_ptr<SubtaskGenerator> subtask_generator : subtask_generators) {
        SharedTasks subtasks = subtask_generator->get_subtasks(task);
        build_abstractions(subtasks, timer, should_abort);
        if (hacked_num_landmark_abstractions == -1) {
            hacked_num_landmark_abstractions = abstractions.size();
        }
        if (should_abort())
            break;
    }
    print_statistics();
}

vector<unique_ptr<Abstraction>> CostSaturation::extract_abstractions() {
    return move(abstractions);
}

vector<shared_ptr<TransitionSystem>> CostSaturation::extract_transition_systems() {
    return move(transition_systems);
}

void CostSaturation::transfer_back_abstractions(std::vector<std::unique_ptr<Abstraction>>&& abstractions)
{
    this->abstractions.clear();
    for (unsigned i = 0; i < abstractions.size(); i++) {
        this->abstractions.emplace_back(std::move(abstractions[i]));
    }
}

void CostSaturation::reset(const TaskProxy &task_proxy) {
    remaining_costs = get_operator_costs(task_proxy);
    num_abstractions = 0;
    num_states = 0;
}

shared_ptr<AbstractTask> CostSaturation::get_remaining_costs_task(
    shared_ptr<AbstractTask> &parent) const {
    vector<int> costs = remaining_costs;
    return make_shared<extra_tasks::ModifiedOperatorCostsTask>(
        parent, move(costs));
}

bool CostSaturation::initial_state_is_dead_end() const {
    return any_of(abstractions.begin(), abstractions.end(),
                  [](const unique_ptr<Abstraction> &abstraction) {
            return abstraction->get_h_value_of_initial_state() == INF;
        });
}

void CostSaturation::build_abstractions(
    const vector<shared_ptr<AbstractTask>> &subtasks,
    const utils::CountdownTimer &timer,
    function<bool()> should_abort) {
    int rem_subtasks = subtasks.size();
    for (shared_ptr<AbstractTask> subtask : subtasks) {
        subtask = get_remaining_costs_task(subtask);

        assert(num_states < max_states);
        unique_ptr<Abstraction> abstraction = utils::make_unique_ptr<Abstraction>(
            subtask,
            max(1, (max_states - num_states) / rem_subtasks),
            max(1, (max_non_looping_transitions - num_non_looping_transitions) /
                rem_subtasks),
            timer.get_remaining_time() / rem_subtasks,
            use_general_costs,
            pick_split,
            *rng,
            compress_cegar_states);

        ++num_abstractions;
        num_states += abstraction->get_num_states();
        assert(num_states <= max_states);
        num_non_looping_transitions += abstraction->get_num_non_looping_transitions();

        if (cost_partitioning_type == CostPartitioningType::SATURATED) {
            reduce_costs(remaining_costs, abstraction->get_saturated_costs());
        }

        if (cost_partitioning_type == CostPartitioningType::SATURATED ||
            cost_partitioning_type == CostPartitioningType::SATURATED_POSTHOC ||
            cost_partitioning_type == CostPartitioningType::SATURATED_MAX) {
            int init_h = abstraction->get_h_value_of_initial_state();
            // transition_systems.push_back(
            //     make_shared<TransitionSystem>(move(*abstraction)));
            if (!exclude_abstractions_with_zero_init_h || init_h > 0) {
                abstractions.push_back(move(abstraction));
            }
        } else if (cost_partitioning_type == CostPartitioningType::OPTIMAL) {
            transition_systems.push_back(
                make_shared<TransitionSystem>(move(*abstraction)));
        } else {
            ABORT("Invalid cost partitioning type");
        }
        if (should_abort())
            break;

        --rem_subtasks;
    }
}

void CostSaturation::print_statistics() const {
    g_log << "Done initializing additive Cartesian heuristic" << endl;
    cout << "Cartesian abstractions built: " << num_abstractions << endl;
    cout << "Abstractions stored: " << abstractions.size() << endl;
    cout << "Transition systems stored: " << transition_systems.size() << endl;
    cout << "Cartesian states: " << num_states << endl;
    cout << "Total number of non-looping transitions: "
         << num_non_looping_transitions << endl;
    cout << endl;
}


void CostSaturation::add_options_to_parser(options::OptionParser& parser)
{
    parser.add_list_option<shared_ptr<SubtaskGenerator>>(
        "subtasks",
        "subtask generators",
        "[landmarks(order=random, random_seed=0, combine_facts=false), goals(order=random, random_seed=0)]");
    parser.add_option<int>(
        "max_states",
        "maximum sum of abstract states over all abstractions",
        "infinity",
        Bounds("1", "infinity"));
    parser.add_option<int>(
        "max_transitions",
        "maximum sum of real transitions (excluding self-loops) over "
        " all abstractions",
        "1000000",
        Bounds("0", "infinity"));
    parser.add_option<double>(
        "max_time",
        "maximum time in seconds for building abstractions",
        "infinity",
        Bounds("0.0", "infinity"));
    /*
      We reserve some memory to be able to recover from out-of-memory
      situations gracefully. When the memory runs out, we stop refining and
      start the next refinement or the search. Due to memory fragmentation
      the memory used for building the abstraction (states, transitions,
      etc.) often can't be reused for things that require big continuous
      blocks of memory. It is for this reason that we require such a large
      amount of memory padding.
    */
    parser.add_option<int>(
        "extra_memory_padding",
        "amount of extra memory in MB to reserve for recovering from"
        " out-of-memory situations gracefully (75 MB work well for CEGAR)",
        "75",
        Bounds("0", "infinity"));
    vector<string> pick_strategies;
    pick_strategies.push_back("RANDOM");
    pick_strategies.push_back("MIN_UNWANTED");
    pick_strategies.push_back("MAX_UNWANTED");
    pick_strategies.push_back("MIN_REFINED");
    pick_strategies.push_back("MAX_REFINED");
    pick_strategies.push_back("MIN_HADD");
    pick_strategies.push_back("MAX_HADD");
    parser.add_enum_option(
        "pick", pick_strategies, "split-selection strategy", "MAX_REFINED");
    vector<string> cp_types;
    cp_types.push_back("SATURATED");
    cp_types.push_back("SATURATED_POSTHOC");
    cp_types.push_back("SATURATED_MAX");
    cp_types.push_back("OPTIMAL");
    parser.add_enum_option(
        "cost_partitioning", cp_types, "cost partitioning method", "SATURATED_POSTHOC");
    parser.add_option<bool>(
        "use_general_costs",
        "allow negative costs in cost partitioning",
        "true");
    parser.add_option<bool>(
        "exclude_abstractions_with_zero_init_h",
        "throw away abstractions with h(s_0) = 0",
        "true");
    parser.add_option<bool>(
        "compress_cartesian_states",
        "throw away abstractions with h(s_0) = 0",
        "true");
    utils::add_rng_options(parser);
}

std::shared_ptr<CostSaturation>
CostSaturation::create(const options::Options& opts,
                       const std::shared_ptr<AbstractTask> task)
{
    const CostPartitioningType cost_partitioning_type =
        static_cast<CostPartitioningType>(opts.get_enum("cost_partitioning"));
    const int extra_memory_padding_in_mb = opts.get<int>("extra_memory_padding");
    const bool exclude_abstractions_with_zero_init_h = opts.get<bool>(
        "exclude_abstractions_with_zero_init_h");
    shared_ptr<utils::RandomNumberGenerator> rng = utils::parse_rng_from_options(opts);

    vector<shared_ptr<SubtaskGenerator>> subtask_generators =
        opts.get_list<shared_ptr<SubtaskGenerator>>("subtasks");
    std::shared_ptr<CostSaturation> cost_saturation = std::make_shared<CostSaturation>(
        cost_partitioning_type,
        subtask_generators,
        opts.get<int>("max_states"),
        opts.get<int>("max_transitions"),
        opts.get<double>("max_time"),
        opts.get<bool>("use_general_costs"),
        exclude_abstractions_with_zero_init_h,
        static_cast<PickSplit>(opts.get<int>("pick")),
        rng,
        opts.get<bool>("compress_cartesian_states"));
    if (extra_memory_padding_in_mb > 0)
        utils::reserve_extra_memory_padding(extra_memory_padding_in_mb);
    cost_saturation->initialize(task);
    if (utils::extra_memory_padding_is_reserved())
        utils::release_extra_memory_padding();
    return cost_saturation;
}

}

static std::shared_ptr<cegar::CostSaturation>
_parse(options::OptionParser& parser)
{
    cegar::CostSaturation::add_options_to_parser(parser);
    ::add_cost_type_option_to_parser(parser);
    parser.add_option<std::shared_ptr<AbstractTask>>(
        "transform",
        "Optional task transformation for the heuristic. "
        "Currently only adapt_costs is available.",
        OptionParser::NONE);
    Options opts = parser.parse();
    if (!parser.dry_run()) {
        std::shared_ptr<AbstractTask> task = get_task_from_options(opts);
        return cegar::CostSaturation::create(opts, task);
    }
    return nullptr;
}

static PluginTypePlugin<cegar::CostSaturation> _plugin_type("CostSaturation", "");
static PluginShared<cegar::CostSaturation> _plugin("cegar", _parse);
