#include "additive_backward_hyperabstraction_heuristic.h"

#include "../option_parser.h"
#include "../plugin.h"
#include "../global_state.h"
#include "../utils/timer.h"
#include "../task_tools.h"

namespace hyperabstraction {

AdditiveBackwardHyperabstractionHeuristic::AdditiveBackwardHyperabstractionHeuristic(
        const options::Options& opts)
    : Heuristic(opts)
    , c_recompute_cost_partioning(opts.get<bool>("dynamic"))
    , m_generator(get_hyperabstraction_generator_from_options(opts))
    , m_cost_partitioning_finder(opts.get<std::shared_ptr<CostPartitioningFinder> >("cost_partitioning"))
    , m_abstraction(nullptr)
{
    std::cout << "Initializing additive backward hyperabstraction heuristic..." << std::endl;
    ::verify_no_axioms(task_proxy); 
    ::verify_no_conditional_effects(task_proxy); 

    HyperabstractionGeneratorResult<LabeledHyperabstraction> r(
            get_labeled_hyperabstraction_from_options(opts, m_generator, false));
    m_abstraction = r.abstraction;
    //r.statistics.print();

    if (!c_recompute_cost_partioning) {
        utils::Timer t;
        std::cout << "Computing static cost partitioning ..." << std::endl;
        m_collection = std::move(m_cost_partitioning_finder->compute_cost_partitioning(
                        *m_abstraction,
                        r.initial_nodes,
                        r.goal_nodes));
        std::cout << "Cost partition collection computation: " << t << std::endl;
        std::cout << "Additive cost partitionings: " << m_collection.get_num_partitions() << std::endl;
        std::cout << "Cost functions: " << m_collection.get_num_cost_functions() << std::endl;
    }

    m_sink_nodes = std::move(r.goal_nodes);
}

int
AdditiveBackwardHyperabstractionHeuristic::compute_heuristic(
        const GlobalState& state)
{
    m_generator->get_nodes(m_source_nodes, state);
    int res = 0;
    if (c_recompute_cost_partioning) {
        res = m_cost_partitioning_finder->compute_cost_partitioning_and_get_distance(
                *m_abstraction,
                m_source_nodes,
                m_sink_nodes);
    } else {
        m_collection.update_distances(*m_abstraction, m_source_nodes);
        res = m_collection.get_max_sum(m_sink_nodes);
    }
    m_source_nodes.clear();
    return res == INF ? DEAD_END : res;
}

    void
AdditiveBackwardHyperabstractionHeuristic::add_options_to_parser(
        options::OptionParser& parser)
{
    add_hyperabstraction_options_to_parser(parser);
    parser.add_option<std::shared_ptr<CostPartitioningFinder> >("cost_partitioning");
    parser.add_option<bool>("dynamic", "", "false");
    Heuristic::add_options_to_parser(parser);
}

}

static Plugin<Heuristic> _plugin(
        "additive_hyb",
        options::parse<Heuristic, hyperabstraction::AdditiveBackwardHyperabstractionHeuristic>);
