#include "additive_forward_hyperabstraction_heuristic.h"

#include "cost_partitioning.h"
#include "utils.h"

#include "../option_parser.h"
#include "../plugin.h"
#include "../global_state.h"
#include "../utils/timer.h"
#include "../task_tools.h"

namespace hyperabstraction {

AdditiveForwardHyperabstractionHeuristic::AdditiveForwardHyperabstractionHeuristic(
        const options::Options& opts)
    : Heuristic(opts)
    , m_generator(get_hyperabstraction_generator_from_options(opts))
{
    std::cout << "Initializing additive forward hyperabstraction heuristic..." << std::endl;
    ::verify_no_axioms(task_proxy); 
    ::verify_no_conditional_effects(task_proxy); 

    HyperabstractionGeneratorResult<LabeledHyperabstraction> r(
            get_labeled_hyperabstraction_from_options(opts, m_generator, true));

    utils::Timer t;
    std::cout << "Computing cost partitioning ..." << std::endl;
    std::shared_ptr<CostPartitioningFinder> cpf =
        opts.get<std::shared_ptr<CostPartitioningFinder> >("cost_partitioning");
    m_collection = std::move(
            cpf->compute_cost_partitioning(*r.abstraction, r.initial_nodes, r.goal_nodes));
    std::cout << "Cost partition collection computation: " << t << std::endl;
    std::cout << "Additive cost partitionings: " << m_collection.get_num_partitions() << std::endl;
    std::cout << "Cost functions: " << m_collection.get_num_cost_functions() << std::endl;

    t.reset();
    std::cout << "Precomputing distances..." << std::endl;
    m_collection.update_distances(*r.abstraction, r.initial_nodes);
    std::cout << "Computed distances in " << t << std::endl;
}

int
AdditiveForwardHyperabstractionHeuristic::compute_heuristic(
        const GlobalState& state)
{
    m_generator->get_nodes(m_sink_nodes, state);
    int res = m_collection.get_max_sum(m_sink_nodes);
    m_sink_nodes.clear();
    return res == INF ? DEAD_END : res;
}

void
AdditiveForwardHyperabstractionHeuristic::add_options_to_parser(
        options::OptionParser& parser)
{
    add_hyperabstraction_options_to_parser(parser);
    parser.add_option<std::shared_ptr<CostPartitioningFinder> >("cost_partitioning");
    Heuristic::add_options_to_parser(parser);
}

}

static Plugin<Heuristic> _plugin(
        "additive_hyf",
        options::parse<Heuristic, hyperabstraction::AdditiveForwardHyperabstractionHeuristic>);
