#include "canonical_cost_partitioning_finder.h"

#include "../cegar/cost_saturation.h"
#include "../cegar/abstraction.h"
#include "../pdbs/pattern_generator.h"
#include "../algorithms/max_cliques.h"

#include "../task_proxy.h"
#include "../abstract_task.h"

#include "../option_parser.h"
#include "../plugin.h"

#include <algorithm>

namespace hyperabstraction {

CanonicalCostPartitioningFinder::CanonicalCostPartitioningFinder(const options::Options& opts)
    : CostPartitioningFinder(opts)
{
    std::vector<std::vector<bool> > compatible;
    std::vector<std::vector<int> > lookup(m_task->get_num_operators());
    std::vector<std::vector<int> > used_operators;
    int num_abstractions = 0;
    if (opts.contains("patterns")) {
        std::shared_ptr<pdbs::PatternCollectionGenerator> gen =
            opts.get<std::shared_ptr<pdbs::PatternCollectionGenerator> >("patterns");
        pdbs::PatternCollectionInformation pattern_collection_info =
            gen->generate(m_task);
        std::shared_ptr<pdbs::PatternCollection> patterns =
            pattern_collection_info.get_patterns();
        num_abstractions = patterns->size();
        compatible.resize(num_abstractions, std::vector<bool>(num_abstractions, true));
        used_operators.resize(num_abstractions);
        std::vector<int> pre(m_task->get_num_variables(), -1);
        std::vector<int> post(m_task->get_num_variables(), -1);
        for (int op = m_task->get_num_operators() - 1; op >= 0; op--) {
            std::fill(pre.begin(), pre.end(), -1);
            std::fill(post.begin(), post.end(), -1);
            for (int i = m_task->get_num_operator_effects(op, false) - 1; i >= 0; i--) {
                FactPair p = m_task->get_operator_effect(op, i, false);
                post[p.var] = p.value;
            }
            for (int i = m_task->get_num_operator_preconditions(op, false) - 1; i >= 0; i--) {
                FactPair p = m_task->get_operator_precondition(op, i, false);
                pre[p.var] = p.value;
                if (post[p.var] == -1) {
                    post[p.var] = p.value;
                }
            }
            std::vector<int>& relevant_patterns = lookup[op];
            for (unsigned p = 0; p < patterns->size(); p++) {
                bool relevant = false;
                const auto& pattern = patterns->at(p);
                for (int i = pattern.size() - 1; i >= 0; i--) {
                    if (pre[pattern[i]] != post[pattern[i]]) {
                        relevant = true;
                        break;
                    }
                }
                if (relevant) {
                    // std::cout 
                    //           << m_task->get_operator_name(op, false)
                    //           << " -> " << p << " ~ " << m_task->get_fact_name(FactPair(pattern[0], 0))
                    //           << std::endl;
                    used_operators[p].push_back(op);
                    for (int i = relevant_patterns.size() - 1; i >= 0; i--) {
                        compatible[relevant_patterns[i]][p] = false;
                    }
                    relevant_patterns.push_back(p);
                }
            }
        }
    }

    if (opts.contains("cartesian")) {
        std::shared_ptr<cegar::CostSaturation> cartesian =
            opts.get<std::shared_ptr<cegar::CostSaturation>>("cartesian");
        std::vector<std::unique_ptr<cegar::Abstraction> > abstractions =
            cartesian->extract_abstractions();
        int offset = num_abstractions;
        num_abstractions += abstractions.size();
        for (int i = compatible.size() - 1; i >= 0; i--) {
            compatible[i].resize(num_abstractions, true);
        }
        compatible.resize(num_abstractions, std::vector<bool>(num_abstractions, true));
        used_operators.resize(num_abstractions);
        std::vector<bool> active;
        for (unsigned i = 0; i < abstractions.size(); i++) {
            int aid = offset + i;
            active = abstractions[i]->compute_active_operators();
            std::vector<int>& ops = used_operators[aid];
            for (int op = active.size() - 1; op >= 0; op--) {
                if (active[op]) {
                    ops.push_back(op);
                    std::vector<int>& relevant = lookup[op];
                    for (int j = relevant.size() - 1; j >= 0; j--) {
                        compatible[relevant[j]][aid] = false;
                    }
                    lookup[op].push_back(aid);
                }
            }
        }
    }

    std::vector<std::vector<int> > cgraph(num_abstractions);
    for (int a1 = 0; a1 < num_abstractions; a1++) {
        for (int a2 = a1 + 1; a2 < num_abstractions; a2++) {
            if (compatible[a1][a2]) {
                cgraph[a1].push_back(a2);
                cgraph[a2].push_back(a1);
            }
        }
    }

    std::vector<std::vector<int> > max_cliques;
    max_cliques::compute_max_cliques(cgraph, max_cliques);
    
    int num_ops = m_task->get_num_operators();
    std::vector<int> costs(num_ops);
    for (int op = num_ops - 1; op >= 0; op--) {
        costs[op] = m_task->get_operator_cost(op, false);
    }

    // for (int i = max_cliques.size() - 1; i >= 0; i--) {
    //     std::sort(max_cliques[i].begin(), max_cliques[i].end());
    // }
    // auto vless = [](const std::vector<int>& x,
    //                 const std::vector<int>& y) {
    //     if (x.size() < y.size()) {
    //         return true;
    //     } else if (x.size() > y.size()) {
    //         return false;
    //     }
    //     for (int i = x.size() - 1; i >= 0; i--) {
    //         if (x[i] < y[i]) {
    //             return true;
    //         } else if (x[i] > y[i]) {
    //             return false;
    //         }
    //     }
    //     return false;
    // };
    // std::sort(max_cliques.begin(), max_cliques.end(), vless);
    // max_cliques.erase(std::unique(max_cliques.begin(), max_cliques.end()), max_cliques.end());

    std::vector<int> mapping(cgraph.size(), -1);
    for (unsigned i = 0; i < max_cliques.size(); i++) {
        std::vector<int>& clique = max_cliques[i];
        m_partitions.m_collection.emplace_back(clique.size());
        for (unsigned j = 0; j < clique.size(); j++) {
            if (mapping[clique[j]] == -1) {
                mapping[clique[j]] = m_partitions.m_cost_partitions.size();
                const std::vector<int>& ops = used_operators[clique[j]];
                m_partitions.m_cost_partitions.emplace_back(num_ops + 1, 0);
                std::vector<int>& c = m_partitions.m_cost_partitions.back();
                for (int k = ops.size() - 1; k >= 0; k--) {
                    c[ops[k]] = costs[ops[k]];
                }
            }
            m_partitions.m_collection.back()[j] = mapping[clique[j]];
        }
    }
}

CostPartitioningCollection
CanonicalCostPartitioningFinder::compute_cost_partitioning(
    LabeledHyperabstraction& ,
    const std::vector<Node>& ,
    const std::vector<Node>& )
{
    return std::move(m_partitions);
}

void 
CanonicalCostPartitioningFinder::add_options_to_parser(options::OptionParser& parser)
{
    parser.add_option<std::shared_ptr<cegar::CostSaturation> >("cartesian",
            "",
            options::OptionParser::NONE);
    parser.add_option<std::shared_ptr<pdbs::PatternCollectionGenerator> >(
            "patterns",
            "",
            options::OptionParser::NONE);
            // "systematic(pattern_max_size=1, only_interesting_patterns=true)");
    CostPartitioningFinder::add_options_to_parser(parser);
}

}

static PluginShared<hyperabstraction::CostPartitioningFinder> _plugin(
        "canonical",
        options::parse_shared<hyperabstraction::CostPartitioningFinder,
                              hyperabstraction::CanonicalCostPartitioningFinder>);

