#ifndef CARTESIAN_HYPERABSTRACTION_GENERATOR_H
#define CARTESIAN_HYPERABSTRACTION_GENERATOR_H

#include "hyperabstraction.h"
#include "hyperabstraction_generator.h"
#include "generic_hyperabstraction_generator.h"

#include "../cegar/domains.h"
#include "../cegar/refinement_hierarchy.h"
#include "../cegar/abstraction.h"


namespace hyperabstraction {


class CartesianNodeMapper {
public:
    CartesianNodeMapper(const std::shared_ptr<AbstractTask> task,
                const cegar::Abstraction& abstraction,
                unsigned offfset);
    template<typename Callback>
    void for_each_intersecting_node(std::vector<int>& head,
                                    std::vector<int>& tail,
                                    Callback& callback) const;
    bool get_compatible_nodes(std::vector<Node>& result,
                                const GlobalState& partial_state,
                                unsigned max_nodes = (unsigned) -1) const;
    bool get_compatible_nodes(std::vector<Node>& result,
                                const std::vector<int>& partial_state,
                                unsigned max_nodes = (unsigned) -1) const;
    bool get_compatible_nodes(std::vector<Node>& result,
                                const cegar::Domains& partial_state,
                                unsigned max_nodes = (unsigned) -1) const;
    cegar::Domains compute_intersection(const Node& node,
                                        const std::vector<int>& state) const;
    cegar::Domains compute_intersection(const Node& node,
                                        const cegar::Domains& state) const;
    unsigned get_num_compatible_nodes(const std::vector<int>& partial_state,
                                        unsigned max_nodes = (unsigned) -1) const;
    unsigned get_num_compatible_nodes(const cegar::Domains& partial_state,
                                        unsigned max_nodes = (unsigned) -1) const;
    unsigned num_nodes() const;
private:
    template<typename State>
    bool get_compatible_nodes_vector(std::vector<Node>& result,
                                     const State& partial_state,
                                     unsigned max_nodes = (unsigned) -1) const;
    template<typename IsDefined, typename NumDefined, typename Callback>
    bool traverse(const cegar::Node* node,
                    const IsDefined& is_defined,
                    const NumDefined& num_defined,
                    Callback& callback) const;
    template<typename IsDefined, typename Callback>
    bool traverse_wrapper(const cegar::Node* node,
                    const IsDefined& is_defined,
                    Callback& callback,
                    std::vector<int>& setminus) const;
    const std::shared_ptr<AbstractTask> m_task;
    const unsigned m_offset;
    const unsigned m_num_nodes;
    std::vector<cegar::Domains> m_cartesian_states;
    const std::shared_ptr<cegar::RefinementHierarchy> m_state_lookup;
};


class CartesianHyperabstractionGenerator : public HyperabstractionGenerator {
public:
    CartesianHyperabstractionGenerator(const options::Options& opts);

    virtual HyperabstractionGeneratorResult<Hyperabstraction> generate_backward() override;
    virtual HyperabstractionGeneratorResult<LabeledHyperabstraction> generate_backward_labeled() override;
    virtual HyperabstractionGeneratorResult<UpdatableHyperabstraction> generate_backward_updatable() override;
    virtual HyperabstractionGeneratorResult<Hyperabstraction> generate_forward() override;
    virtual HyperabstractionGeneratorResult<LabeledHyperabstraction> generate_forward_labeled() override;
    virtual HyperabstractionGeneratorResult<UpdatableHyperabstraction> generate_forward_updatable() override;

    virtual void get_nodes(std::vector<Node>& nodes, const GlobalState& state) override;
    virtual unsigned num_nodes() const override;

    static void add_options_to_parser(options::OptionParser& parser);

private:
    template<typename Abstraction, bool Direction>
    HyperabstractionGeneratorResult<Abstraction> generate();

    std::vector<CartesianNodeMapper> m_node_mappers;
};


template<typename Abstraction, bool Direction>
HyperabstractionGeneratorResult<Abstraction>
CartesianHyperabstractionGenerator::generate()
{
    return generate_hyperabstraction<Abstraction>(
            m_task,
            Direction,
            c_tau_max_nodes,
            c_tau_max_nodes_goal,
            c_cache_tau_nodes,
            c_transition_pruning,
            false,
            c_transition_enumeration_depth,
            c_goal_enumeration_depth,
            c_enumeration_max_nodes,
            m_node_mappers);
}


}




namespace hyperabstraction {

template<typename Callback>
void
CartesianNodeMapper::for_each_intersecting_node(
        std::vector<int>& head,
        std::vector<int>& tail,
        Callback& callback) const
{
    auto is_defined = [&head](const int& var, const int& val) {
        return head[var] == -1 || head[var] == val;
    };
    auto num_defined = [&](const int& var) {
        return (unsigned) (head[var] == -1 ? m_task->get_variable_domain_size(var) : 1);
    };
    auto wrapper = [&](const Node& node_id) {
        const cegar::Domains& s = m_cartesian_states[node_id];
        cegar::Domains cartesian_state(s);
        for (int var = head.size() - 1; var >= 0; var--) {
            if (tail[var] != -1) {
                cartesian_state.set_single_value(var, tail[var]);
            } else if (head[var] != -1) {
                cartesian_state.add_all(var);
            }
        }
        callback(m_offset + node_id, cartesian_state);
        return true;
    };
    traverse(m_state_lookup->get_root(), 
             is_defined,
             num_defined,
             wrapper);
}

template<typename IsDefined, typename NumDefined, typename Callback>
bool
CartesianNodeMapper::traverse(
        const cegar::Node* node,
        const IsDefined& is_defined,
        const NumDefined& num_defined,
        Callback& callback) const
{
    static std::vector<int> setminus(m_task->get_num_variables());
    for (int var = setminus.size() - 1; var >= 0; var--) {
        setminus[var] = num_defined(var);
    }
    return traverse_wrapper(node, is_defined, callback, setminus);
}

template<typename IsDefined, typename Callback>
bool
CartesianNodeMapper::traverse_wrapper(
        const cegar::Node* node,
        const IsDefined& is_defined,
        Callback& callback,
        std::vector<int>& setminus) const
{
    if (!node->is_split()) {
        return callback(node->get_state_id());
    } else {
        int var = node->get_var();
        int not_matched = setminus[var];
        const cegar::Node* left = node;
        const cegar::Node* right = node->get_right_child();
                                                                                // std::cout << "  - branch(var=" << var << ", num=" << num << ") - " << std::endl;
        while (left->is_split() && left->get_right_child() == right) {
            assert(left->get_var() == var);
                                                                                // std::cout << "   -> check (var=" << var << ", val=" << left->get_val() << " => ";
            if (is_defined(var, left->get_val())) {
                if (--not_matched == 0) {
                                                                                // std::cout << num << std::endl;
                    break;
                }
            }
                                                                                // std::cout << num << std::endl;
            left = left->get_left_child();
        }

        int num = setminus[var];

        bool result = true;
        if (not_matched > 0) {
                                                                                // if (r && num > 0) std::cout << "   ~~left~~" << std::endl;
            setminus[var] = not_matched;
            result = traverse_wrapper(left, is_defined, callback, setminus);
        }

        if (result && not_matched != num) {
                                                                                // if (sat) std::cout << "   ~~right~~" << std::endl;
            setminus[var] = num - not_matched;
            result = traverse_wrapper(right, is_defined, callback, setminus);
        }

                                                                                // std::cout << "  - backtrack(var=" << var << ") - " << std::endl;

        setminus[var] = num;
        return result;
    }
}


template<typename State>
bool
CartesianNodeMapper::get_compatible_nodes_vector(
        std::vector<Node>& result,
        const State& partial_state,
        unsigned max_nodes) const
{
    auto is_defined = [&partial_state](const int& var, const int& val) {
        return (partial_state[var] == -1) || (partial_state[var] == val);
    };
    auto num_defined = [&](const int& var) {
        return (unsigned) (partial_state[var] == -1 ? m_task->get_variable_domain_size(var) : 1);
    };
    unsigned previous = result.size();
    max_nodes += previous;
    auto callback = [&](const Node& node) {
        if (result.size() == max_nodes) {
            return false;
        }
        assert(std::count(result.begin(), result.end(), m_offset+node) == 0);
        result.push_back(m_offset + node);
        return true;
    };
    if (traverse(m_state_lookup->get_root(), is_defined, num_defined, callback)) {
#ifndef NDEBUG
        for (unsigned i = previous; i < result.size(); i++) {
            for (int var = 0; var < m_task->get_num_variables(); var++) {
                assert(partial_state[var] == -1
                        || m_cartesian_states[result[i] - m_offset].test(var, partial_state[var]));
            }
        }
#endif
        return true;
    }
    result.resize(previous);
    return false;
}


}

#endif
