#include "cegar_cost_partition.h"

#include "../option_parser.h"
#include "../plugin.h"

#include "../cegar/cost_saturation.h"
#include "../cegar/abstraction.h"
#include "../cegar/additive_cartesian_heuristic.h"
#include "../cegar/scp_optimizer.h"
#include "../task_tools.h"

#include "cost_partitioning.h"

using namespace std;
using namespace cegar;

namespace hyperabstraction {


static vector<CostPartitioning> compute_all_saturated_cost_partitionings(
    const vector<unique_ptr<Abstraction>> &abstractions,
    const vector<int> &operator_costs) {
    vector<int> indices = get_default_order(abstractions.size());

    std::vector<CostPartitioning> cost_partitionings;
    do {
        vector<vector<int>> saturated_costs;
            compute_saturated_cost_partitioning(
                abstractions, indices, operator_costs, saturated_costs);

        cost_partitionings.push_back(std::move(saturated_costs));
    } while (next_permutation(indices.begin(), indices.end()));

    return std::move(cost_partitionings);
}

CegarCostPartition::CegarCostPartition(const options::Options & opts) :
    CostPartitioningFinder(opts), cost_partitioning(compute_cegar_partitioning(opts, m_task)) {
    for (int i = cost_partitioning.size() - 1; i >= 0; i--) {
        for (int j = cost_partitioning[i].size() - 1; j >= 0; j--) {
            cost_partitioning[i][j].push_back(0);
        }
    }

}

std::vector<CostPartitioning> CegarCostPartition::compute_cegar_partitioning (const options::Options& opts,
                                                                std::shared_ptr<AbstractTask> m_task) {

    std::shared_ptr<cegar::CostSaturation> cegar =
        opts.get<std::shared_ptr<cegar::CostSaturation> >("abstraction");

    vector<unique_ptr<cegar::Abstraction>> abstractions =
        cegar->extract_abstractions();

    if (cegar->get_cost_partitioning_type() == CostPartitioningType::OPTIMAL) {
        cerr << "CostPartitioningType::OPTIMAL not supported in hyperabstraction::CegarCostPartition" << endl;
        utils::exit_with(utils::ExitCode::INPUT_ERROR);
    }

    assert(cegar->get_cost_partitioning_type() == CostPartitioningType::SATURATED ||
            cegar->get_cost_partitioning_type() == CostPartitioningType::SATURATED_POSTHOC ||
            cegar->get_cost_partitioning_type() == CostPartitioningType::SATURATED_MAX);

    const int num_orders = opts.get<int>("orders");
    const int max_num_samples = opts.get<int>("samples");
    const double max_sampling_time = opts.get<double>("max_sampling_time");
    const double max_optimization_time = opts.get<double>("max_optimization_time");
    const double max_time_finding_orders = opts.get<double>("max_time_finding_orders");
    const bool shuffle = opts.get<bool>("shuffle");
    const bool reverse_order = opts.get<bool>("reverse");
    const bool diversify = opts.get<bool>("diversify");

    if (num_orders > 1 && !shuffle) {
        cerr << "When using more than one order set shuffle=true" << endl;
        utils::exit_with(utils::ExitCode::INPUT_ERROR);
    }
    if (reverse_order && (num_orders > 1 || shuffle)) {
        cerr << "Use reverse=true only with shuffle=false and orders=1" << endl;
        utils::exit_with(utils::ExitCode::INPUT_ERROR);
    }
    if (diversify && num_orders == 1) {
        cerr << "When diversifying set orders > 1" << endl;
        utils::exit_with(utils::ExitCode::INPUT_ERROR);
    }
    if (diversify && !shuffle) {
        cerr << "When diversifying set shuffle=true" << endl;
        utils::exit_with(utils::ExitCode::INPUT_ERROR);
    }
    if (diversify && max_num_samples == 0) {
        cerr << "When diversifying set samples >= 1" << endl;
        utils::exit_with(utils::ExitCode::INPUT_ERROR);
    }
    if (max_num_samples == 0 && max_optimization_time != numeric_limits<double>::infinity()) {
        cerr << "Option max_optimization_time has no effect when samples == 0" << endl;
        utils::exit_with(utils::ExitCode::INPUT_ERROR);
    }

    utils::Timer independence_timer;
    vector<vector<bool>> dependent_ops;
    for (const unique_ptr<Abstraction> &abstraction : abstractions) {
        dependent_ops.push_back(abstraction->compute_active_operators());
    }

    int num_pairs = 0;
    int num_independent_pairs = 0;

    for (size_t i = 0; i < dependent_ops.size(); ++i) {
        for (size_t j = i + 1; j < dependent_ops.size(); ++j) {
            if (disjunct(dependent_ops[i], dependent_ops[j])) {
                ++num_independent_pairs;
            }
            ++num_pairs;
        }
    }

    cout << num_independent_pairs << "/" << num_pairs << " = "
            << (num_pairs ? num_independent_pairs * 100. / num_pairs : 100.)
            << "% of abstraction pairs are independent" << endl;
    cout << "Time for computing independence: " << independence_timer << endl;

    vector<shared_ptr<RefinementHierarchy>> refinement_hierarchies;
    refinement_hierarchies.reserve(abstractions.size());
    for (unique_ptr<Abstraction> &abstraction : abstractions) {
        refinement_hierarchies.push_back(abstraction->get_refinement_hierarchy());
    }

    if (cegar->get_cost_partitioning_type() == CostPartitioningType::SATURATED) {
        std::vector<std::vector<int> > saturated_costs;
        for (const unique_ptr<Abstraction> &abstraction : abstractions) {
            saturated_costs.push_back(abstraction->get_saturated_costs());
            bool zero = true;
            const auto& costs = saturated_costs.back();
            // std::cout << "[";
            for (int i = costs.size() - 1; i >= 0; i--) {
                // std::cout << ((i < (int) costs.size() - 1) ? ", " : "")
                //           << costs[i];
                zero = zero && (costs[i] <= 0);
                if (!zero) break;
            }
            // std::cout << "] -> " << zero << std::endl;
            if (zero) {
                saturated_costs.pop_back();
            }
        }

        return std::vector<CostPartitioning>({std::move(saturated_costs)});
    }

    TaskProxy task_proxy (*m_task);
    vector<int> operator_costs = get_operator_costs(task_proxy);

    if (cegar->get_cost_partitioning_type() == CostPartitioningType::SATURATED_MAX) {
        return compute_all_saturated_cost_partitionings(abstractions, operator_costs);
    }

    vector<vector<int>> saturated_costs_default_order;
    vector<vector<int>> h_values_by_abstraction_for_default_order =
        compute_saturated_cost_partitioning(
            abstractions, get_default_order(abstractions.size()), operator_costs, saturated_costs_default_order);

    State initial_state = task_proxy.get_initial_state();
    vector<int> local_state_ids_for_initial_state = get_local_state_ids(refinement_hierarchies, initial_state);
    int init_h = compute_sum_h(
        local_state_ids_for_initial_state,
        h_values_by_abstraction_for_default_order);
    cout << "Initial h value for default order: " << init_h << endl;
    if (init_h == INF) {
        return std::vector<CostPartitioning>({std::move(saturated_costs_default_order)});
    }

    auto rng = cegar->get_rng();

    std::vector<CostPartitioning> cost_partitionings;

    SCPOptimizer scp_optimizer(move(abstractions), refinement_hierarchies, operator_costs, rng);
    vector<vector<int>> local_state_ids_by_state =
        sample_states_and_return_local_ids(
            task_proxy, m_task,
            refinement_hierarchies,
            h_values_by_abstraction_for_default_order,
            init_h,
            max_num_samples,
            max_sampling_time,
            *rng);
    int num_samples = local_state_ids_by_state.size();

    utils::CountdownTimer finding_orders_timer(max_time_finding_orders);
    int total_num_evaluated_orders = 0;
    vector<vector<vector<int>>> h_values_by_orders;
    vector<int> portfolio_h_values(num_samples, 0);
    vector<int> portfolio_h_values_improvement(num_samples, -1);

    utils::Timer update_portfolio_h_values_timer;
    update_portfolio_h_values_timer.stop();

    for (int i = 0; i < num_orders && !finding_orders_timer.is_expired(); ++i) {
        double optimization_time = min(
            max_optimization_time, finding_orders_timer.get_remaining_time());
        std::vector<std::vector<int>> saturated_costs;
        pair<vector<vector<int>>, pair<int, int>> result =
            scp_optimizer.find_cost_partitioning(
                local_state_ids_by_state,
                optimization_time,
                shuffle,
                reverse_order,
                portfolio_h_values,
                portfolio_h_values_improvement, saturated_costs);

        vector<vector<int>> h_values_by_abstraction = move(result.first);
        int total_h_value = result.second.first;
        int num_evaluated_orders = result.second.second;
        total_num_evaluated_orders += num_evaluated_orders;
        if (!diversify || total_h_value > 0 || h_values_by_orders.empty()) {
            if (diversify) {
                update_portfolio_h_values_timer.resume();
                update_portfolio_h_values(
                    portfolio_h_values,
                    portfolio_h_values_improvement,
                    local_state_ids_by_state);
                update_portfolio_h_values_timer.stop();
            }
            h_values_by_orders.push_back(move(h_values_by_abstraction));
            cost_partitionings.push_back(std::move(saturated_costs));
        }
    }

    cout << "Total evaluated orders: " << total_num_evaluated_orders << endl;
    cout << "Orders: " << h_values_by_orders.size() << endl;

    cout << "Time for computing SCPs: "
            << *scp_optimizer.scp_computation_timer << endl;
    cout << "Time for evaluating orders: "
            << *scp_optimizer.order_evaluation_timer << endl;
    cout << "Time for updating portfolio h values: "
            << update_portfolio_h_values_timer << endl;
    cout << "Time for finding orders: " << finding_orders_timer << endl;

    cegar->transfer_back_abstractions(std::move(abstractions));

    return std::move(cost_partitionings);
}



// std::vector<std::vector<int> >  CegarCostPartition::get_single_cost_partitioning(
//     LabeledHyperabstraction& ,
//     const std::vector<Node>& ,
//     const std::vector<Node>& ) {
//     std::cout << "CegarCostPartition::get_single_cost_partitioning not implemented" << std::endl;
//     utils::exit_with(utils::ExitCode::UNSUPPORTED);
// }

void CegarCostPartition::add_options_to_parser(options::OptionParser& parser){

    parser.add_option<std::shared_ptr<cegar::CostSaturation> >("abstraction",
                                                                "",
                                                                "cegar(compress_cegar_states=false)");

    parser.add_option<bool>(
        "use_general_costs",
        "allow negative costs in cost partitioning",
        "true");
    parser.add_option<int>(
        "orders",
        "number of abstraction orders to maximize over",
        "1",
        Bounds("1", "infinity"));
    parser.add_option<int>(
        "samples",
        "number of sample states to optimize for",
        "0",
        Bounds("0", "infinity"));
    parser.add_option<double>(
        "max_sampling_time",
        "maximum time for finding samples",
        "infinity",
        Bounds("0.0", "infinity"));
    parser.add_option<double>(
        "max_optimization_time",
        "maximum time in seconds for optimizing each order",
        "infinity",
        Bounds("0.0", "infinity"));
    parser.add_option<double>(
        "max_time_finding_orders",
        "maximum total time in seconds for finding all orders",
        "infinity",
        Bounds("0.0", "infinity"));
    parser.add_option<bool>(
        "shuffle",
        "shuffle order before optimizing it",
        "true");
    parser.add_option<bool>(
        "reverse",
        "reverse order before optimizing it (to obtain hadd-up order)",
        "false");
    parser.add_option<bool>(
        "diversify",
        "search for orders that complement the portfolio",
        "false");


    hyperabstraction::CostPartitioningFinder::add_options_to_parser(parser);

}
}

static std::shared_ptr<hyperabstraction::CostPartitioningFinder> _parse(options::OptionParser& parser)
{
hyperabstraction::CegarCostPartition::add_options_to_parser(parser);

options::Options opts = parser.parse();
if (!parser.dry_run()) {
    return std::make_shared<hyperabstraction::CegarCostPartition>(opts);
}
return nullptr;
}

static PluginShared<hyperabstraction::CostPartitioningFinder> _plugin(
"cegar",
_parse);


