#include "combined_hyperabstraction_generator.h"

#include "../option_parser.h"
#include "../plugin.h"

#include "../global_state.h"
#include "../pdbs/pattern_generator.h"
#include "../cegar/cost_saturation.h"

namespace hyperabstraction {

CombinedHyperabstractionGenerator::CombinedHyperabstractionGenerator(const options::Options& opts)
    : HyperabstractionGenerator(opts)
{
    unsigned num = 0;
    if (opts.contains("patterns")) {
        std::shared_ptr<pdbs::PatternCollectionGenerator> gen =
            opts.get<std::shared_ptr<pdbs::PatternCollectionGenerator> >("patterns");
        std::shared_ptr<pdbs::PatternCollection> collection
            = gen->generate(m_task).get_patterns();
        m_pdbs.reserve(collection->size());
        for (unsigned i = 0; i < collection->size(); i++) {
            m_pdbs.emplace_back(m_task,
                                collection->at(i),
                                num,
                                opts.get<bool>("enumerate_tail"));
            num += m_pdbs.back().num_nodes();
        }
        std::cout << "PDB states: " << num << std::endl;
    }
    if (opts.contains("cartesian")) {
        unsigned pdb = num;
        std::shared_ptr<cegar::CostSaturation> cegar =
            opts.get<std::shared_ptr<cegar::CostSaturation> >("cartesian");
        std::vector<std::unique_ptr<cegar::Abstraction> > abstractions =
            cegar->extract_abstractions();
        m_cartesians.reserve(abstractions.size());
        for (unsigned i = 0; i < abstractions.size(); i++) {
            m_cartesians.emplace_back(m_task, *abstractions[i], num);
            num += m_cartesians.back().num_nodes();
        }
        cegar->transfer_back_abstractions(std::move(abstractions));
        std::cout << "Cartesian states: " << (num - pdb) << std::endl;
    }
}

HyperabstractionGeneratorResult<Hyperabstraction>
CombinedHyperabstractionGenerator::generate_backward()
{
    return generate<Hyperabstraction, false>();
}

HyperabstractionGeneratorResult<LabeledHyperabstraction>
CombinedHyperabstractionGenerator::generate_backward_labeled()
{
    return generate<LabeledHyperabstraction, false>();
}

HyperabstractionGeneratorResult<UpdatableHyperabstraction>
CombinedHyperabstractionGenerator::generate_backward_updatable()
{
    return generate<UpdatableHyperabstraction, false>();
}

HyperabstractionGeneratorResult<Hyperabstraction>
CombinedHyperabstractionGenerator::generate_forward()
{
    return generate<Hyperabstraction, true>();
}

HyperabstractionGeneratorResult<LabeledHyperabstraction>
CombinedHyperabstractionGenerator::generate_forward_labeled()
{
    return generate<LabeledHyperabstraction, true>();
}

HyperabstractionGeneratorResult<UpdatableHyperabstraction>
CombinedHyperabstractionGenerator::generate_forward_updatable()
{
    return generate<UpdatableHyperabstraction, true>();
}

void
CombinedHyperabstractionGenerator::get_nodes(std::vector<Node>& nodes,
                                        const GlobalState& state)
{
    for (int i = m_pdbs.size() - 1; i >= 0; i--) {
        nodes.push_back(m_pdbs[i].get_node(state));
    }
    for (int i = m_cartesians.size() - 1; i >= 0; i--) {
        m_cartesians[i].get_compatible_nodes(nodes, state);
    }
}

unsigned
CombinedHyperabstractionGenerator::num_nodes() const
{
    unsigned res = 0;
    for (int i = m_pdbs.size() - 1; i >= 0; i--) {
        res += m_pdbs[i].num_nodes();
    }
    for (int i = m_cartesians.size() - 1; i >= 0; i--) {
        res += m_cartesians[i].num_nodes();
    }
    return res;
}


void
CombinedHyperabstractionGenerator::add_options_to_parser(options::OptionParser& parser)
{
    parser.add_option<std::shared_ptr<pdbs::PatternCollectionGenerator> >(
            "patterns",
            "",
            options::OptionParser::NONE);
    parser.add_option<std::shared_ptr<cegar::CostSaturation> >(
            "cartesian",
            "",
            options::OptionParser::NONE);
    parser.add_option<bool>("enumerate_tail", "", "false");
    HyperabstractionGenerator::add_options_to_parser(parser);
}

}

static PluginShared<hyperabstraction::HyperabstractionGenerator> _plugin(
        "combined",
        options::parse_shared<hyperabstraction::HyperabstractionGenerator, hyperabstraction::CombinedHyperabstractionGenerator>);
