#include "cost_partitioning.h"
#include "../utils/hash.h"
#include <algorithm>
#include <unordered_set>

namespace hyperabstraction {

CostPartitioningCollection::CostPartitioningCollection()
{}

CostPartitioningCollection::CostPartitioningCollection(
        std::vector<CostPartitioning>&& collection)
    : m_collection(collection.size())
{
    std::unordered_set<int, utils::CollectionIdHasher<CostPartitioning>, utils::CollectionIdComparator<CostPartitioning> > set(
            collection.size() * collection.front().size(),
            utils::CollectionIdHasher<CostPartitioning>(m_cost_partitions),
            utils::CollectionIdComparator<CostPartitioning>(m_cost_partitions));
    using Collection = std::vector<unsigned>;
    std::unordered_set<int, utils::CollectionIdHasher<std::vector<Collection> >, utils::CollectionIdComparator<std::vector<Collection> > > collection_set(
            collection.size(),
            utils::CollectionIdHasher<std::vector<Collection> >(m_collection),
            utils::CollectionIdComparator<std::vector<Collection> >(m_collection));
    unsigned k = 0;
    for (unsigned i = 0; i < collection.size(); i++) {
        CostPartitioning& p = collection[i];
        std::vector<unsigned>& references = m_collection[k];
        for (unsigned j = 0; j < p.size(); j++) {
            m_cost_partitions.emplace_back(std::move(p[j]));
            bool keep = false;
            for (int op = m_cost_partitions.back().size() - 1; op >= 0; op--) {
                if (m_cost_partitions.back()[op] > 0) {
                    keep = true;
                    break;
                }
            }
            if (!keep) {
                m_cost_partitions.pop_back();
                continue;
            }
            auto x = set.insert(m_cost_partitions.size() - 1);
            if (x.second) {
                references.push_back(m_cost_partitions.size() - 1);
            } else {
                references.push_back(*x.first);
                m_cost_partitions.pop_back();
            }
        }
        std::sort(references.begin(), references.end());
        auto x = collection_set.insert(k);
        if (x.second) {
            k++;
        } else {
            references.clear();
        }
    }
    m_collection.resize(k);
    m_distances.resize(m_cost_partitions.size());
}

CostPartitioningCollection& CostPartitioningCollection::operator=(
        std::vector<CostPartitioning>&& collection)
{
    return ((*this) = std::move(CostPartitioningCollection(std::move(collection))));
}

CostPartitioningCollection::CostPartitioningCollection(
        const CostPartitioning& cps,
        const std::vector<std::vector<unsigned> >& collection)
    : m_cost_partitions(cps)
    , m_collection(collection)
    , m_distances(m_cost_partitions.size())
{}

unsigned
CostPartitioningCollection::get_num_partitions() const
{
    return m_collection.size();
}

unsigned
CostPartitioningCollection::get_num_cost_functions() const
{
    return m_cost_partitions.size();
}

void
CostPartitioningCollection::update_distances(
        LabeledHyperabstraction& abstraction,
        const std::vector<Node>& from)
{
    m_distances.resize(m_cost_partitions.size());
    for (int i = m_cost_partitions.size() - 1; i >= 0; i--) {
        m_distances[i].resize(abstraction.num_nodes());
        abstraction.set_costs(m_cost_partitions[i]);
        abstraction.compute_distances(m_distances[i], from);
    }
}

int
CostPartitioningCollection::get_max_sum(
        const std::vector<Node>& nodes) const
{
    static std::vector<int> max_values(m_distances.size());
    assert(m_distances.size() == m_cost_partitions.size());
    assert(max_values.size() == m_cost_partitions.size());
    for (int i = m_distances.size() - 1; i >= 0; i--) {
        max_values[i] = get_max_distance(m_distances[i], nodes);
        if (max_values[i] == INF) {
            return INF;
        }
    }
    int max_sum = 0;
    for (int i = m_collection.size() - 1; i >= 0; i--) {
        const std::vector<unsigned>& references = m_collection[i];
        int sum = 0;
        for (int j = references.size() - 1; j >= 0; j--) {
            sum += max_values[references[j]];
        }
        if (sum > max_sum) {
            max_sum = sum;
        }
    }
    return max_sum;
}

}
