#ifdef GENERIC_HYPERABSTRACTION_GENERATOR_H

#include "hyperabstraction_transition_generator.h"

#include "../abstract_task.h"
#include "../utils/hash.h"
#include "../segmented_vector.h"

#include <vector>
#include <unordered_map>
#include <algorithm>
#include <cassert>
#include <limits>
#include <tuple>
#include <type_traits>

namespace hyperabstraction {

template<typename Hyperabstraction>
HyperabstractionGeneratorResult<Hyperabstraction>::HyperabstractionGeneratorResult(
        HyperabstractionGeneratorStatistics&& statistics,
        std::shared_ptr<Hyperabstraction>&& ptr,
        std::vector<Node>&& initial_nodes,
        std::vector<Node>&& goal_nodes,
        std::vector<int>&& distances)
    : statistics(std::move(statistics))
    , abstraction(std::move(ptr))
    , initial_nodes(std::move(initial_nodes))
    , goal_nodes(std::move(goal_nodes))
    , distances(std::move(distances))
{}

template<typename Hyperabstraction,
         typename TransitionGenerator,
         typename TauNodeGetter,
         typename Head,
         typename Tail,
         typename... AbstractionCollection>
class GenericHyperabstractionGenerator {
public:
    GenericHyperabstractionGenerator(
        const std::shared_ptr<AbstractTask> task,
        unsigned t_tau_max_nodes,
        unsigned g_tau_max_nodes,
        TauNodeGetter& get_tau_node,
        bool prune_inf,
        Head& get_head,
        Tail& get_tail,
        const std::vector<int>& initial_state,
        const std::vector<int>& goal_state,
        unsigned transition_enumeration_depth,
        unsigned goal_enumeration_depth,
        unsigned enumeration_max_nodes,
        HyperabstractionGeneratorStatistics&& statistics,
        AbstractionCollection&... collection)
        : statistics(std::move(statistics))
        , collections(std::forward_as_tuple(collection...))
        , iterate_collections(
                std::make_tuple<IterateIncreasingNumCompatibleNodes<AbstractionCollection>... >(collection...))
        , num_nodes(get_num_nodes())
        , result(std::make_shared<Hyperabstraction>(num_nodes, task->get_num_operators() + 1))
        , transition_generator(*result, this->statistics)
        , initial_state(initial_state)
        , goal_state(goal_state)
        , task(task)
        , tau_max_nodes(t_tau_max_nodes)
        , t_tau_max_nodes(t_tau_max_nodes)
        , g_tau_max_nodes(g_tau_max_nodes)
        , get_tau_node(get_tau_node)
        , get_head(get_head)
        , get_tail(get_tail)
        , tau_label(task->get_num_operators())
        , prune_inf(prune_inf)
        , transition_enumeration_depth(transition_enumeration_depth)
        , goal_enumeration_depth(goal_enumeration_depth)
        , enumeration_max_nodes(enumeration_max_nodes)
    {
    }

    HyperabstractionGeneratorResult<Hyperabstraction>
    get()
    {
        construct();
        return HyperabstractionGeneratorResult<Hyperabstraction>(
                std::move(statistics),
                std::move(result),
                std::move(initial_nodes),
                std::move(goal_nodes),
                std::move(distances));
    }

private:
    void construct();
    template<bool Enumerative>
    void setup_transitions();
    void setup_initial_and_goal_nodes();
    void run_postprocessing();
    template<typename Abstraction, typename AbstractState>
    bool get_compatible_nodes_create_tau1(std::vector<Node>& nodes,
                                          Abstraction& base_abs,
                                          const AbstractState& tail_state);
    template<typename Callback, typename AbstractState>
    void get_compatible_nodes_enumerative(const unsigned& depth_left,
                                          const AbstractState& state,
                                          Callback& callback,
                                          unsigned collectionid = -1,
                                          int idx = -1,
                                          Node head = no_node);
    template<typename Callback, typename AbstractState>
    void get_compatible_nodes_enumerative(std::vector<Node>& in_construction,
                                          unsigned depth_left,
                                          const AbstractState& state,
                                          const Node& exclude_node,
                                          Callback& callback);

    template<typename AbstractState>
    void generate_transitions_simple(const Node& head_node,
                                     const AbstractState& tail_state,
                                     const int& label,
                                     const int& cost);

    template<std::size_t I = 0>
    typename std::enable_if<I == sizeof...(AbstractionCollection), unsigned>::type
    get_num_nodes() { return 0; }

    template<std::size_t I = 0>
    typename std::enable_if<I < sizeof...(AbstractionCollection), unsigned>::type
    get_num_nodes()
    {
        unsigned num = 0;
        auto& collection = std::get<I>(collections);
        for (int i = collection.size() - 1; i >= 0; i--) {
            num += collection[i].num_nodes();
        }
        return num + get_num_nodes<I+1>();
    }

    template<std::size_t I = 0, typename AbstractState>
    typename std::enable_if<I == sizeof...(AbstractionCollection), void>::type
    get_compatible_nodes(std::vector<Node>& , const AbstractState& ) {}

    template<std::size_t I = 0, typename AbstractState>
    typename std::enable_if<I < sizeof...(AbstractionCollection), void>::type
    get_compatible_nodes(std::vector<Node>& tail_nodes, const AbstractState& tail_state)
    {
        auto& collection = std::get<I>(collections);    
        for (int idx = collection.size() - 1; idx >= 0; idx--) {
            collection[idx].get_compatible_nodes(tail_nodes, tail_state, -1);
        }
        get_compatible_nodes<I+1>(tail_nodes, tail_state);
    }

    template<std::size_t I = 0, typename AbstractState>
    typename std::enable_if<I == sizeof...(AbstractionCollection), void>::type
    get_compatible_nodes_create_tau(std::vector<Node>&, const AbstractState& ) {}

    template<std::size_t I = 0, typename AbstractState>
    typename std::enable_if<I < sizeof...(AbstractionCollection), void>::type
    get_compatible_nodes_create_tau(
            std::vector<Node>& tail_nodes,
            const AbstractState& tail_state)
    {
        auto& collection = std::get<I>(collections);
        for (int idx = collection.size() - 1; idx >= 0; idx--) {
            get_compatible_nodes_create_tau1(tail_nodes, collection[idx], tail_state);
        }
        get_compatible_nodes_create_tau<I+1>(tail_nodes, tail_state);
    }

    template<std::size_t I = 0>
    typename std::enable_if<I == sizeof...(AbstractionCollection), void>::type
    generate_transitions_enumerative(const int&, const int&, std::vector<int>&, std::vector<int>&) {}

    template<std::size_t I = 0>
    typename std::enable_if<I < sizeof...(AbstractionCollection), void>::type
    generate_transitions_enumerative(
            const int& op,
            const int& cost,
            std::vector<int>& head,
            std::vector<int>& tail)
    {
        auto& collection = std::get<I>(collections);
        for (int idx = collection.size() - 1; idx >= 0; idx--) {
            auto temp = [&](const Node& node, const auto& tail_state) {
                auto add_t = [this, &node, &op, &cost](const std::vector<Node>& tail) {
                    transition_generator(node, tail, op, cost);
                };
                get_compatible_nodes_enumerative(transition_enumeration_depth,
                                                 tail_state,
                                                 add_t,
                                                 I,
                                                 idx,
                                                 node);
            };
            collection[idx].for_each_intersecting_node(head, tail, temp);
        }
        generate_transitions_enumerative<I+1>(op, cost, head, tail);
    }

    template<std::size_t I = 0>
    typename std::enable_if<I == sizeof...(AbstractionCollection), void>::type
    generate_transitions_simple(const int&, const int&, std::vector<int>&, std::vector<int>&) {}

    template<std::size_t I = 0>
    typename std::enable_if<I < sizeof...(AbstractionCollection), void>::type
    generate_transitions_simple(
            const int& op,
            const int& cost,
            std::vector<int>& head,
            std::vector<int>& tail)
    {
        auto& collection = std::get<I>(collections);
        for (int idx = collection.size() - 1; idx >= 0; idx--) {
            auto temp = [this, &op, &cost](const Node& node, const auto& tail_state) {
                generate_transitions_simple(node, tail_state, op, cost);
            };
            collection[idx].for_each_intersecting_node(head, tail, temp);
        }
        generate_transitions_simple<I+1>(op, cost, head, tail);
    }

    template<std::size_t I = 0, typename AbstractState>
    typename std::enable_if<I == sizeof...(AbstractionCollection), void>::type
    initialize_iterators(const unsigned&, const AbstractState&, const unsigned&, const int& ) { }

    template<std::size_t I = 0, typename AbstractState>
    typename std::enable_if<I < sizeof...(AbstractionCollection), void>::type
    initialize_iterators(const unsigned& tau, const AbstractState& state, const unsigned& i, const int& idx)
    {
        auto& it = std::get<I>(iterate_collections);
        it.update(tau, state, i == I ? idx : -1);
        initialize_iterators<I+1>(tau, state, i, idx);
    }

    template<std::size_t I = 0, typename AbstractState>
    typename std::enable_if<I == sizeof...(AbstractionCollection), bool>::type
    get_remaining_compatible(std::vector<Node>&, const AbstractState&, const Node&) { return true; }

    template<std::size_t I = 0, typename AbstractState>
    typename std::enable_if<I < sizeof...(AbstractionCollection), bool>::type
    get_remaining_compatible(std::vector<Node>& compatible, const AbstractState& state, const Node& exclude)
    {
        auto& it = std::get<I>(iterate_collections);
        for (int i = it.left() - 1; i >= 0; i--) {
            auto& abstraction = it.get(i);
            if (get_compatible_nodes_create_tau1(compatible, abstraction, state)
                    && compatible.back() == exclude) {
                return false;
            }
        }
        return get_remaining_compatible<I+1>(compatible, state, exclude);
    }

    template<std::size_t I = 0>
    typename std::enable_if<I == sizeof...(AbstractionCollection), unsigned>::type
    get_best_value_for_next_iterator() { return (unsigned) -1;  }

    template<std::size_t I = 0>
    typename std::enable_if<I < sizeof...(AbstractionCollection), unsigned>::type
    get_best_value_for_next_iterator()
    {
        unsigned best = get_best_value_for_next_iterator<I+1>();
        unsigned cur  = std::get<I>(iterate_collections).value();
        return best < cur ? best : cur;
    }

    template<std::size_t I = 0, typename AbstractState, typename Callback>
    typename std::enable_if<I == sizeof...(AbstractionCollection), bool>::type
    enumerate(std::vector<Node>&, unsigned , const AbstractState& , const Node& , Callback& , const unsigned& ) { return false; }

    template<std::size_t I = 0, typename AbstractState, typename Callback>
    typename std::enable_if<I < sizeof...(AbstractionCollection), bool>::type
    enumerate(std::vector<Node>& set,
              unsigned depth_left,
              const AbstractState& state,
              const Node& exclude,
              Callback& callback,
              const unsigned& best)
    {
        assert(depth_left > 0);
        auto& it = std::get<I>(iterate_collections);
        if (it.left() == 0 || it.value() > best) {
            return enumerate<I+1>(set, depth_left, state, exclude, callback, best);
        }
        auto& abstraction = it.get();
        std::vector<Node> compatible;
        if (abstraction.get_compatible_nodes(compatible, state, enumeration_max_nodes)) {
            assert(!compatible.empty());
            it.next();
            for (int i = compatible.size() - 1; i >= 0; i--) {
                if (compatible[i] == exclude) {
                    continue;
                }
                set.push_back(compatible[i]);
                get_compatible_nodes_enumerative(set,
                                                 depth_left - 1,
                                                 abstraction.compute_intersection(compatible[i], state),
                                                 exclude,
                                                 callback);
                set.pop_back();
            }
            it.previous();
            return true;
        } else {
            return false;
        }
    }

    template<typename Collection>
    struct IterateIncreasingNumCompatibleNodes {
        IterateIncreasingNumCompatibleNodes(Collection& collection);

        template<typename AbstractState>
        void update(const unsigned& maxtaus,
                    const AbstractState& state,
                    const int& force_first);

        unsigned value() const;
        auto& get() { assert(m_next >= 0); return m_collection[m_ordered_indxs[m_next].second]; }
        auto& get(const int& i) { assert(i >= 0 && i <= m_next); return m_collection[m_ordered_indxs[i].second]; }
        unsigned left() const { return m_next + 1; }
        void next() { m_next--; }
        void previous() { m_next++; }

    private:

        struct Comparator {
            bool operator()(const std::pair<unsigned, unsigned>& x,
                            const std::pair<unsigned, unsigned>& y) const
            {
                return x.first > y.first;
            }
        };

        Collection& m_collection;

        Comparator m_comp;
        std::vector<std::pair<unsigned, unsigned> > m_ordered_indxs;
        int m_next;
    };

    HyperabstractionGeneratorStatistics statistics;

    std::tuple<AbstractionCollection&...> collections;
    std::tuple<IterateIncreasingNumCompatibleNodes<AbstractionCollection>... > iterate_collections;

    unsigned num_nodes;

    std::shared_ptr<Hyperabstraction> result;
    TransitionGenerator transition_generator;

    const std::vector<int>& initial_state;
    const std::vector<int>& goal_state;
    const std::shared_ptr<AbstractTask> task;
    unsigned tau_max_nodes;
    unsigned t_tau_max_nodes;
    unsigned g_tau_max_nodes;
    TauNodeGetter& get_tau_node;
    Head& get_head;
    Tail& get_tail;
    int tau_label;
    bool prune_inf;
    unsigned transition_enumeration_depth;
    unsigned goal_enumeration_depth;
    unsigned enumeration_max_nodes;

    std::vector<Node> initial_nodes;
    std::vector<Node> goal_nodes;
    std::vector<int> distances;

};

template<typename Hyperabstraction,
         typename TransitionGenerator,
         typename TauNodeGetter,
         typename Head,
         typename Tail,
         typename... AbstractionCollection>
template<typename Collection>
GenericHyperabstractionGenerator<Hyperabstraction, TransitionGenerator, TauNodeGetter, Head, Tail, AbstractionCollection...>
::IterateIncreasingNumCompatibleNodes<Collection>::IterateIncreasingNumCompatibleNodes(
        Collection& collection)
    : m_collection(collection)
    , m_ordered_indxs(collection.size())
    , m_next(-1)
{
}

template<typename Hyperabstraction,
         typename TransitionGenerator,
         typename TauNodeGetter,
         typename Head,
         typename Tail,
         typename... AbstractionCollection>
template<typename Collection>
template<typename AbstractState>
void
GenericHyperabstractionGenerator<Hyperabstraction, TransitionGenerator, TauNodeGetter, Head, Tail, AbstractionCollection...>
::IterateIncreasingNumCompatibleNodes<Collection>::update(
        const unsigned& max_nodes,
        const AbstractState& state,
        const int& force_first)
{
    unsigned m_max_nodes = (max_nodes == (unsigned) -1) ? max_nodes : (max_nodes + 1);
    int j = 0;
    for (int i = m_collection.size() - 1; i >= 0; i--) {
        m_ordered_indxs[j].second = i;
        if (i == force_first) {
            m_ordered_indxs[j].first = 0;
        } else {
            m_ordered_indxs[j].first = m_collection[i].get_num_compatible_nodes(state, m_max_nodes);
            if (m_ordered_indxs[j].first == 0 || m_ordered_indxs[j].first == m_collection[i].num_nodes()) {
                assert(m_ordered_indxs[j].first > 0);
                j--;
            }
        }
        j++;
    }
    std::sort(m_ordered_indxs.begin(), m_ordered_indxs.begin() + j, m_comp);
    m_next = j - 1;
}

template<typename Hyperabstraction,
         typename TransitionGenerator,
         typename TauNodeGetter,
         typename Head,
         typename Tail,
         typename... AbstractionCollection>
template<typename Collection>
unsigned
GenericHyperabstractionGenerator<Hyperabstraction, TransitionGenerator, TauNodeGetter, Head, Tail, AbstractionCollection...>
::IterateIncreasingNumCompatibleNodes<Collection>::value() const
{
    return m_next < 0 ? (unsigned) -1 : m_ordered_indxs[m_next].first;
}

template<typename Hyperabstraction,
         typename TransitionGenerator,
         typename TauNodeGetter,
         typename Head,
         typename Tail,
         typename... AbstractionCollection>
void
GenericHyperabstractionGenerator<Hyperabstraction, TransitionGenerator, TauNodeGetter, Head, Tail, AbstractionCollection...>
::construct()
{
    statistics.begin_construction();

    if (transition_enumeration_depth > 0) {
        setup_transitions<true>();
    } else {
        setup_transitions<false>();
    }
    setup_initial_and_goal_nodes();
    run_postprocessing();

    statistics.end_construction(result->num_nodes(),
                                result->num_transitions(),
                                result->num_hypertransitions_strict());
}

template<typename Hyperabstraction,
         typename TransitionGenerator,
         typename TauNodeGetter,
         typename Head,
         typename Tail,
         typename... AbstractionCollection>
template<bool Enumerative>
void
GenericHyperabstractionGenerator<Hyperabstraction, TransitionGenerator, TauNodeGetter, Head, Tail, AbstractionCollection...>
::setup_transitions()
{
    std::vector<int> pre (task->get_num_variables());
    std::vector<int> post(task->get_num_variables());
    for (int opidx = task->get_num_operators() - 1; opidx >= 0; opidx--) {
        std::fill(pre.begin(), pre.end(), -1);
        std::fill(post.begin(), post.end(), -1);
        for (int j = task->get_num_operator_effects(opidx, false) - 1; j >= 0; j--) {
            FactPair p = task->get_operator_effect(opidx, j, false);
            post[p.var] = p.value;
        }
        for (int j = task->get_num_operator_preconditions(opidx, false) - 1; j >= 0; j--) {
            FactPair p = task->get_operator_precondition(opidx, j, false);
            pre[p.var] = p.value;
            if (post[p.var] == -1) {
                post[p.var] = p.value;
            }
        }
        std::vector<int>& head = *(get_head(pre, post));
        std::vector<int>& tail = *(get_tail(pre, post));
        int cost = task->get_operator_cost(opidx, false);
        if (Enumerative) {
            generate_transitions_enumerative(opidx, cost, head, tail);
        } else {
            generate_transitions_simple(opidx, cost, head, tail);
        }
    }
}

template<typename Hyperabstraction,
         typename TransitionGenerator,
         typename TauNodeGetter,
         typename Head,
         typename Tail,
         typename... AbstractionCollection>
void
GenericHyperabstractionGenerator<Hyperabstraction, TransitionGenerator, TauNodeGetter, Head, Tail, AbstractionCollection...>
::setup_initial_and_goal_nodes()
{

    get_compatible_nodes(initial_nodes, initial_state);
    // std::cout << "I = [";
    // std::sort(initial_nodes.begin(), initial_nodes.end());
    // for (unsigned i = 0; i < initial_nodes.size(); i++) {
    //     std::cout << (i > 0 ? ", " : "") << initial_nodes[i];
    // }
    // std::cout << "]" << std::endl;

    tau_max_nodes = g_tau_max_nodes;

    bool fully_defined = true;
    for (int var = goal_state.size() - 1; var >= 0; var--) {
        if (goal_state[var] == -1) {
            fully_defined = false;
            break;
        }
    }

    if (fully_defined || goal_enumeration_depth == 0) {
        get_compatible_nodes_create_tau(goal_nodes, goal_state);
    } else {
        Node goal_node = result->num_nodes();
        result->add_nodes(1);
        statistics.inc_tau_nodes(1);
        auto add_to_goal = [&](const std::vector<Node>& tail) {
            result->add_transition(goal_node, tail, tau_label, 0);
            statistics.inc_tau_transitions(1);
        };
        goal_nodes.push_back(goal_node);

        get_compatible_nodes_enumerative(goal_enumeration_depth,
                                         goal_state,
                                         add_to_goal);
                                         
    }

    // std::sort(goal_nodes.begin(), goal_nodes.end());
    // std::cout << "G = [";
    // for (unsigned i = 0; i < goal_nodes.size(); i++) {
    //     std::cout << (i > 0 ? ", " : "") << goal_nodes[i];
    // }
    // std::cout << "]" << std::endl;

    tau_max_nodes = t_tau_max_nodes;
}

template<typename Hyperabstraction,
         typename TransitionGenerator,
         typename TauNodeGetter,
         typename Head,
         typename Tail,
         typename... AbstractionCollection>
void
GenericHyperabstractionGenerator<Hyperabstraction, TransitionGenerator, TauNodeGetter, Head, Tail, AbstractionCollection...>
::run_postprocessing()
{
    // if (prune_inf) {
    //     distances.resize(result->num_nodes(), INF);
    //     result->compute_distances(distances, initial_nodes);
    //     unsigned ts_before = result->num_transitions();
    //     result->prune_unreachable();
    //     statistics.inc_dinf_transitions(ts_before - result->num_transitions());
    //     for (int i = distances.size() - 1; i >= 0; i--) {
    //         if (distances[i] == INF) {
    //             statistics.inc_dinf_nodes();
    //         }
    //     }
    // }
}

template<typename Hyperabstraction,
         typename TransitionGenerator,
         typename TauNodeGetter,
         typename Head,
         typename Tail,
         typename... AbstractionCollection>
template<typename Abstraction, typename AbstractState>
bool
GenericHyperabstractionGenerator<Hyperabstraction, TransitionGenerator, TauNodeGetter, Head, Tail, AbstractionCollection...>
::get_compatible_nodes_create_tau1(std::vector<Node>& tail_nodes,
                                   Abstraction& base_abs,
                                   const AbstractState& tail_state)
{
    static std::vector<Node> tau_nodes;
    bool added = true;
    if (base_abs.get_compatible_nodes(tau_nodes, tail_state, tau_max_nodes)
            && tau_nodes.size() < base_abs.num_nodes()) {
        if (tau_nodes.size() <= 1) {
            tail_nodes.insert(tail_nodes.end(), tau_nodes.begin(), tau_nodes.end());
            added = tau_nodes.size() > 0;
        } else {
            Node tau_node = get_tau_node(tau_nodes, result->num_nodes());
            if (tau_node == result->num_nodes()) {
                result->add_nodes(1);
                for (int i = tau_nodes.size() - 1; i >= 0; i--) {
                    result->add_transition(tau_node, std::vector<Node>({tau_nodes[i]}), tau_label, 0);
                }
                statistics.inc_tau_nodes();
                statistics.inc_tau_transitions(tau_nodes.size());
            }
            tail_nodes.push_back(tau_node);
        }
    }
    tau_nodes.clear();
    return added;
}

template<typename Hyperabstraction,
         typename TransitionGenerator,
         typename TauNodeGetter,
         typename Head,
         typename Tail,
         typename... AbstractionCollection>
template<typename Callback, typename AbstractState>
void
GenericHyperabstractionGenerator<Hyperabstraction, TransitionGenerator, TauNodeGetter, Head, Tail, AbstractionCollection...>
::get_compatible_nodes_enumerative(const unsigned& depth_left,
                                   const AbstractState& state,
                                   Callback& callback,
                                   unsigned collectionid,
                                   int idx,
                                   Node head)
{
    static std::vector<Node> compatible;
    assert(compatible.empty());
    initialize_iterators(enumeration_max_nodes, state, collectionid, idx);
    get_compatible_nodes_enumerative(compatible,
                                     depth_left,
                                     state,
                                     head,
                                     callback);
    compatible.clear();
}

template<typename Hyperabstraction,
         typename TransitionGenerator,
         typename TauNodeGetter,
         typename Head,
         typename Tail,
         typename... AbstractionCollection>
template<typename Callback, typename AbstractState>
void
GenericHyperabstractionGenerator<Hyperabstraction, TransitionGenerator, TauNodeGetter, Head, Tail, AbstractionCollection...>
::get_compatible_nodes_enumerative(std::vector<Node>& in_construction,
                                   unsigned depth_left,
                                   const AbstractState& state,
                                   const Node& exclude_node,
                                   Callback& callback)
{
    unsigned best = get_best_value_for_next_iterator();
    if (depth_left == 0
            || !enumerate(in_construction, depth_left, state, exclude_node, callback, best)) {
        unsigned size = in_construction.size();
        if (get_remaining_compatible(in_construction, state, exclude_node)) {
            callback(in_construction);
        }
        in_construction.resize(size);
    }
}

template<typename Hyperabstraction,
         typename TransitionGenerator,
         typename TauNodeGetter,
         typename Head,
         typename Tail,
         typename... AbstractionCollection>
template<typename AbstractState>
void
GenericHyperabstractionGenerator<Hyperabstraction, TransitionGenerator, TauNodeGetter, Head, Tail, AbstractionCollection...>
::generate_transitions_simple(const Node& head_node,
                              const AbstractState& tail_state,
                              const int& label,
                              const int& cost)
{
    static std::vector<Node> tail_nodes;
    get_compatible_nodes_create_tau(tail_nodes, tail_state);
    transition_generator(head_node, tail_nodes, label, cost);
    tail_nodes.clear();
}

template<typename Hyperabstraction,
         typename TauNodeGetter,
         typename GetHead,
         typename GetTail,
         typename... AbstractionCollection>
HyperabstractionGeneratorResult<Hyperabstraction>
generate_hyperabstraction2(const std::shared_ptr<AbstractTask> task,
                           bool /*forward*/,
                           unsigned t_tau_max_nodes,
                           unsigned g_tau_max_nodes,
                           TauNodeGetter& tau_node_getter,
                           const TransitionPruningMethod& /*pruning*/,
                           bool prune_inf,
                           GetHead& get_head,
                           GetTail& get_tail,
                           const std::vector<int>& initial_state,
                           const std::vector<int>& goal,
                           unsigned transition_enumeration_depth,
                           unsigned goal_enumeration_depth,
                           unsigned enumeration_max_nodes,
                           HyperabstractionGeneratorStatistics&& statistics,
                           AbstractionCollection&... collection...)
{
    // if (pruning == TransitionPruningMethod::SUBSUMED) {
    //     // if (forward) {
    //     //     return GenericHyperabstractionGenerator<Hyperabstraction,
    //     //                                             TransitionGeneratorPruneSubsumed<Hyperabstraction, true>,
    //     //                                             TauNodeGetter,
    //     //                                             GetHead,
    //     //                                             GetTail,
    //     //                                             AbstractionCollection...>(
    //     //                                                     task,
    //     //                                                     tau_max_nodes,
    //     //                                                     tau_node_getter,
    //     //                                                     prune_inf,
    //     //                                                     get_head,
    //     //                                                     get_tail,
    //     //                                                     initial_state,
    //     //                                                     goal,
    //     //                                                     transition_enumeration_depth,
    //     //                                                     goal_enumeration_depth,
    //     //                                                     enumeration_max_nodes,
    //     //                                                     std::move(statistics),
    //     //                                                     collection...
    //     //                                                     ).get();
    //     // } else {
    //     //     return GenericHyperabstractionGenerator<Hyperabstraction,
    //     //                                             TransitionGeneratorPruneSubsumed<Hyperabstraction, false>,
    //     //                                             TauNodeGetter,
    //     //                                             GetHead,
    //     //                                             GetTail,
    //     //                                             AbstractionCollection...>(
    //     //                                                     task,
    //     //                                                     tau_max_nodes,
    //     //                                                     tau_node_getter,
    //     //                                                     prune_inf,
    //     //                                                     get_head,
    //     //                                                     get_tail,
    //     //                                                     initial_state,
    //     //                                                     goal,
    //     //                                                     transition_enumeration_depth,
    //     //                                                     goal_enumeration_depth,
    //     //                                                     enumeration_max_nodes,
    //     //                                                     std::move(statistics),
    //     //                                                     collection...
    //     //                                                     ).get();
    //     // }
    // }

    // if (pruning == TransitionPruningMethod::DISABLED) {
    //     return GenericHyperabstractionGenerator<Hyperabstraction,
    //                                             TransitionGeneratorNoPrune<Hyperabstraction>,
    //                                             TauNodeGetter,
    //                                             GetHead,
    //                                             GetTail,
    //                                             AbstractionCollection...>(
    //                                                     task,
    //                                                     t_tau_max_nodes,
    //                                                     g_tau_max_nodes,
    //                                                     tau_node_getter,
    //                                                     prune_inf,
    //                                                     get_head,
    //                                                     get_tail,
    //                                                     initial_state,
    //                                                     goal,
    //                                                     transition_enumeration_depth,
    //                                                     goal_enumeration_depth,
    //                                                     enumeration_max_nodes,
    //                                                     std::move(statistics),
    //                                                     collection...
    //                                                     ).get();
    // } 

    return GenericHyperabstractionGenerator<Hyperabstraction,
                                            TransitionGeneratorPruneDuplicate<Hyperabstraction>,
                                            TauNodeGetter,
                                            GetHead,
                                            GetTail,
                                            AbstractionCollection...>(
                                                    task,
                                                    t_tau_max_nodes,
                                                    g_tau_max_nodes,
                                                    tau_node_getter,
                                                    prune_inf,
                                                    get_head,
                                                    get_tail,
                                                    initial_state,
                                                    goal,
                                                    transition_enumeration_depth,
                                                    goal_enumeration_depth,
                                                    enumeration_max_nodes,
                                                    std::move(statistics),
                                                    collection...
                                                    ).get();
}

template<typename Hyperabstraction,
         typename GetHead,
         typename GetTail,
         typename... AbstractionCollection>
HyperabstractionGeneratorResult<Hyperabstraction>
generate_hyperabstraction1(const std::shared_ptr<AbstractTask> task,
                           bool forward,
                           unsigned t_tau_max_nodes,
                           unsigned g_tau_max_nodes,
                           bool /*cache_tau_nodes*/,
                           const TransitionPruningMethod& pruning,
                           bool prune_inf,
                           GetHead& get_head,
                           GetTail& get_tail,
                           const std::vector<int>& initial_state,
                           const std::vector<int>& goal,
                           unsigned transition_enumeration_depth,
                           unsigned goal_enumeration_depth,
                           unsigned enumeration_max_nodes,
                           HyperabstractionGeneratorStatistics&& statistics,
                           AbstractionCollection&... collection...)
{
    // if (cache_tau_nodes) {
        struct TauNodeHashMap {
            TauNodeHashMap() : data(1024, Hasher(sets), Comparator(sets)) {}
            Node operator()(std::vector<Node>& nodes, Node default_value)
            {
                sets.push_back(nodes);
                std::vector<Node>& sat = sets[sets.size() - 1];
                std::sort(sat.begin(), sat.end());
                auto it = data.insert(std::pair<int, Node>(sets.size() - 1, default_value));
                if (it.second) {
                    return default_value;
                }
                sets.resize(sets.size() - 1);
                return it.first->second;
            }
        private:
            using Hasher = utils::CollectionIdHasher<SegmentedVector<std::vector<Node> > >;
            using Comparator = utils::CollectionIdComparator<SegmentedVector<std::vector<Node> > >;
            SegmentedVector<std::vector<Node> > sets;
            std::unordered_map<int, Node, Hasher, Comparator> data;
        };

        TauNodeHashMap tau_node_getter;
        return generate_hyperabstraction2<Hyperabstraction>(
                task,
                forward,
                t_tau_max_nodes,
                g_tau_max_nodes,
                tau_node_getter,
                pruning,
                prune_inf,
                get_head,
                get_tail,
                initial_state,
                goal,
                transition_enumeration_depth,
                goal_enumeration_depth,
                enumeration_max_nodes,
                std::move(statistics),
                collection...);
    // } else {
    //     static auto tau_node_getter = [](const std::vector<Node>& , Node default_value) {
    //         return default_value;
    //     };
    //     return generate_hyperabstraction2<Hyperabstraction>(
    //             task,
    //             forward,
    //             t_tau_max_nodes,
    //             g_tau_max_nodes,
    //             tau_node_getter,
    //             pruning,
    //             prune_inf,
    //             get_head,
    //             get_tail,
    //             initial_state,
    //             goal,
    //             transition_enumeration_depth,
    //             goal_enumeration_depth,
    //             enumeration_max_nodes,
    //             std::move(statistics),
    //             collection...);
    // }
}

template<typename Hyperabstraction, typename... AbstractionCollection>
HyperabstractionGeneratorResult<Hyperabstraction>
generate_hyperabstraction(const std::shared_ptr<AbstractTask> task,
                          bool forward,
                          unsigned t_tau_max_nodes,
                          unsigned g_tau_max_nodes,
                          bool cache_tau_nodes,
                          const TransitionPruningMethod& pruning,
                          bool prune_inf,
                          unsigned transition_enumeration_depth,
                          unsigned goal_enumeration_depth,
                          unsigned enumeration_max_nodes,
                          AbstractionCollection&... collection)
{
    HyperabstractionGeneratorStatistics statistics;
    statistics.set_caching_tau_nodes(cache_tau_nodes);
    statistics.set_transition_pruning_method(pruning);
    statistics.set_tau_max_nodes(t_tau_max_nodes, g_tau_max_nodes);
    statistics.set_prune_inf(prune_inf);
    statistics.begin_construction();

    std::shared_ptr<Hyperabstraction> result = nullptr;

    std::vector<int> initial_state(task->get_initial_state_values());
    std::vector<int> goal(task->get_num_variables(), -1);
    for (int i = task->get_num_goals() - 1; i >= 0; i--) {
        FactPair p = task->get_goal_fact(i);
        goal[p.var] = p.value;
    }

    if (forward) {
        static auto head = [](std::vector<int>& pre, std::vector<int>& ) {
            return &pre;
        };
        static auto tail = [](std::vector<int>&, std::vector<int>& post) {
            return &post;
        };
        return generate_hyperabstraction1<Hyperabstraction>(
                task,
                forward,
                t_tau_max_nodes,
                g_tau_max_nodes,
                cache_tau_nodes,
                pruning,
                prune_inf,
                head,
                tail,
                goal,
                initial_state,
                transition_enumeration_depth,
                goal_enumeration_depth,
                enumeration_max_nodes,
                std::move(statistics),
                collection...);
    } else {
        static auto head = [](std::vector<int>&, std::vector<int>& post) {
            return &post;
        };
        static auto tail = [](std::vector<int>& pre, std::vector<int>& ) {
            return &pre;
        };
        return generate_hyperabstraction1<Hyperabstraction>(
                task,
                forward,
                t_tau_max_nodes,
                g_tau_max_nodes,
                cache_tau_nodes,
                pruning,
                prune_inf,
                head,
                tail,
                initial_state,
                goal,
                transition_enumeration_depth,
                goal_enumeration_depth,
                enumeration_max_nodes,
                std::move(statistics),
                collection...);
    }
}

}

#endif
