#include "graph_cut_cost_partitioning.h"

#include "../abstract_task.h"
#include "../option_parser.h"
#include "../plugin.h"
#include "cost_partitioning.h"

#include <deque>
#include <set>
#include <algorithm>
#include <limits>

namespace hyperabstraction {

    using namespace std;

GraphCutCostPartitioning::GraphCutCostPartitioning(const options::Options& opts)
    : CostPartitioningFinder(opts)
    , m_default_cost(m_task->get_num_operators() + 1, 0)
{
    for (int i = m_task->get_num_operators() - 1; i >= 0; i--) {
        m_default_cost[i] = m_task->get_operator_cost(i, false);
    }
}

std::vector<CostPartitioning>
GraphCutCostPartitioning::get_cost_partitioning(
    LabeledHyperabstraction& abstraction,
    const std::vector<Node>& source,
    const std::vector<Node>& sink)
{
    compute_cost_partitioning_and_get_distance(abstraction, source, sink);
    std::vector<CostPartitioning> result(1);
    CostPartitioning& cp = result[0];
    cp.resize(m_cuts.size(), std::vector<int>(m_task->get_num_operators() + 1, 0));
    for (unsigned i = 0; i < m_cuts.size(); i++) {
        std::vector<int>& costs = cp[i];
        int cost = m_cuts[i].first;
        const auto& cut = m_cuts[i].second;
        for (int j = cut.size() - 1; j >= 0; j--) {
            costs[cut[j]] = cost;
        }
    }
#ifndef NDEBUG
    for (int i = 0; i < m_task->get_num_operators(); i++) {
        int cost = m_task->get_operator_cost(i, false);
        for (int j = cp.size() - 1; j >= 0; j--) {
            cost -= cp[j][i];
        }
        assert(cost >= 0);
    }
#endif
    return std::move(result);
}

int
GraphCutCostPartitioning::compute_cost_partitioning_and_get_distance(
    LabeledHyperabstraction& abstraction,
    const std::vector<Node>& source,
    const std::vector<Node>& sink)
{
    m_cuts.clear();

    std::vector<int> distances(abstraction.num_nodes());
    std::vector<std::vector<std::pair<Node, int> > > graph(abstraction.num_nodes());
    std::vector<int> remainder(m_default_cost);
    std::set<int> cut;
    std::vector<std::vector<int> > result;
    std::vector<bool> in_sink_zone(abstraction.num_nodes(), false);
    std::deque<Node> sink_zone;
    auto on_transition = [&graph](const Node& max_tail,
                                  const PerTransitionInfo<false>& t,
                                  const int& ) {
        assert(t.head < graph.size());
        graph[t.head].emplace_back(max_tail, t.label); 
    };

    int summed = 0;

    while (true) {
        result.emplace_back(remainder.size(), 0);

        graph.clear(); graph.resize(abstraction.num_nodes());
        abstraction.set_costs(remainder);
        abstraction.compute_distances(distances,
                                      source,
                                      on_transition);

        Node max_sink_node = no_node;
        int max_sink_distance = 0;
        for (int i = sink.size() - 1; i >= 0; i--) {
            if (distances[sink[i]] > max_sink_distance) {
                max_sink_node = sink[i];
                max_sink_distance = distances[sink[i]];
            }
        }

        if (max_sink_distance == INF) {
            summed = INF;
            break;
        } else if (max_sink_distance == 0) {
            break;
        }

        assert(cut.empty() && sink_zone.empty());

        int min_cost_cut = std::numeric_limits<int>::max();
        sink_zone.push_back(max_sink_node);
        std::fill(in_sink_zone.begin(), in_sink_zone.end(), false);
        in_sink_zone[max_sink_node] = true;
        while (!sink_zone.empty()) {
            Node n = sink_zone.back();
            sink_zone.pop_back();
            const std::vector<std::pair<Node, int> >& incoming = graph[n];
            for (int i = incoming.size() - 1; i >= 0; i--) {
                const int& op = incoming[i].second;
                if (remainder[op] == 0) {
                    const Node& m = incoming[i].first;
                    if (!in_sink_zone[m]) {
                        sink_zone.push_back(m);
                        in_sink_zone[m] = true;
                    }
                } else {
                    min_cost_cut = std::min(min_cost_cut, remainder[op]);
                    cut.insert(op);
                }
            }
        }

        summed += min_cost_cut;

        m_cuts.emplace_back();
        m_cuts.back().first = min_cost_cut;
        m_cuts.back().second.insert(m_cuts.back().second.end(),
                                    cut.begin(),
                                    cut.end());
        for (auto it = cut.begin(); it != cut.end(); it++) {
            remainder[*it] -= min_cost_cut;
            assert(remainder[*it] >= 0);
        }
        cut.clear();
    }

    std::cout << "h -> " << summed << std::endl;
    return summed;
}

void
GraphCutCostPartitioning::add_options_to_parser(options::OptionParser& parser)
{
    CostPartitioningFinder::add_options_to_parser(parser);
}

}

static PluginShared<hyperabstraction::CostPartitioningFinder> _plugin(
        "graph_cut",
        options::parse_shared<hyperabstraction::CostPartitioningFinder,
        hyperabstraction::GraphCutCostPartitioning>);

