#ifdef HYPERABSTRACTION_H

#include <cassert>

namespace hyperabstraction {


template<>
struct __TransitionValue<true> {
    int value;
    __TransitionValue() : value(0) {}
    void set_value(const int& v) { value = v; }
};

template<bool Updatable>
PerTransitionInfo<Updatable>::PerTransitionInfo(
        const Node& head,
        const int& tail_size,
        const int& cost)
    : tail_size(tail_size)
    , cost(cost)
    , head(head)
    , count(0)
{}





template<bool Updatable>
__HyperabstractionBase<Updatable>::__HyperabstractionBase(unsigned nodes,
                                                          unsigned labels)
    : m_dummy(nodes)
    , m_num_labels(labels)
    , m_has_negative_cost(false)
    , m_had_negative_cost(false)
{
    m_in_tail.resize(nodes + 1);
}

template<bool Updatable>
void
__HyperabstractionBase<Updatable>::set_transition_cost(
        PerTransitionInfo<Updatable>* t,
        const int& cost) 
{
    t->cost = cost;
    if (cost < 0) {
        set_has_negative_cost();
    }
}

template<bool Updatable>
PerTransitionInfo<Updatable>&
__HyperabstractionBase<Updatable>::get_transition(unsigned id) 
{
    assert(id < m_transition_info.size());
    return m_transition_info[id];
}

template<bool Updatable>
const PerTransitionInfo<Updatable>&
__HyperabstractionBase<Updatable>::get_transition(unsigned id) const
{
    assert(id < m_transition_info.size());
    return m_transition_info[id];
}

template<bool Updatable>
const std::vector<PerTransitionInfo<Updatable> >&
__HyperabstractionBase<Updatable>::get_transitions_with_tail(const Node& node) const
{
    assert(node < m_in_tail.size());
    return m_in_tail[node];
}

template<bool Updatable>
unsigned
__HyperabstractionBase<Updatable>::num_nodes() const
{
    return m_in_tail.size();
}

template<bool Updatable>
unsigned
__HyperabstractionBase<Updatable>::num_transitions() const
{
    return m_transition_info.size();
}

template<bool Updatable>
unsigned
__HyperabstractionBase<Updatable>::num_hypertransitions_strict() const
{
    unsigned res = 0;
    for (int i = num_transitions() - 1; i >= 0; i--) {
        if (m_transition_info[i].tail_size > 1) {
            res += 1;
        }
    }
    return res;
}

template<bool Updatable>
unsigned
__HyperabstractionBase<Updatable>::num_labels() const
{
    return m_num_labels;
}

template<bool Updatable>
bool
__HyperabstractionBase<Updatable>::has_negative_cost() const
{
    return m_has_negative_cost;;
}

template<bool Updatable>
void
__HyperabstractionBase<Updatable>::set_has_negative_cost() 
{
    m_has_negative_cost = true;
    if (!m_had_negative_cost) {
        m_had_negative_cost = true;
        // m_transition_tails.resize(this->num_transitions());
        // for (unsigned i = 0; i < m_transition_info.size(); i++) {
        //     m_transition_info[i].count = i;
        // }
        for (int node = m_in_tail.size() - 1; node >= 0; node--) {
            const std::vector<PerTransitionInfo<Updatable>* >& ts = m_in_tail[node];
            for (int i = ts.size() - 1; i >= 0; i--) {
                m_transition_tails[ts[i]].push_back(node);
            }
        }
    }
}





template<bool Updatable, bool Labeled>
__LabeledHyperabstraction<Updatable, Labeled>::__LabeledHyperabstraction(
        unsigned nodes,
        unsigned labels)
    : __HyperabstractionBase<Updatable>(nodes, labels)
{}

template<bool Updatable, bool Labeled>
void
__LabeledHyperabstraction<Updatable, Labeled>::set_costs(const std::vector<int>&)
{
}

template<bool Updatable, bool Labeled>
void
__LabeledHyperabstraction<Updatable, Labeled>::add_transition_label(
        PerTransitionInfo<Updatable>* ,
        const int& )
{
}

template<bool Updatable, bool Labeled>
void
__LabeledHyperabstraction<Updatable, Labeled>::remove_transition_label(
        PerTransitionInfo<Updatable>* ,
        const int& )
{
}

template<bool Updatable, bool Labeled>
void
__LabeledHyperabstraction<Updatable, Labeled>::update_labels_remove_unreachable_transitions()
{
}

template<bool Updatable, bool Labeled>
int
__LabeledHyperabstraction<Updatable, Labeled>::get_label(const PerTransitionInfo<Updatable>* ) const
{
    return -1;
}

template<bool Updatable>
class __LabeledHyperabstraction<Updatable, true> : public __HyperabstractionBase<Updatable> {
public:
    __LabeledHyperabstraction(unsigned nodes,
                              unsigned labels)
        : __HyperabstractionBase<Updatable>(nodes, labels)
        , m_label_to_transitions(labels)
    {
    }

    void set_costs(const std::vector<int>& costs)
    {
        assert(costs.size() == m_label_to_transitions.size());
        this->m_has_negative_cost = false;
        static int max_cost = (INF / 2);
        for (int i = this->num_transitions() - 1; i >= 0; i--) {
            this->m_transition_info[i].cost = max_cost;
        }
        bool neg = false;
        for (int i = costs.size() - 1; i >= 0; i--) {
            const int& c = costs[i];
            neg = neg || c < 0;
            std::vector<PerTransitionInfo<Updatable>*>& transitions = m_label_to_transitions[i];
            for (int j = transitions.size() - 1; j >= 0; j--) {
                PerTransitionInfo<Updatable>* t = transitions[j];
                t->cost = std::min(t->cost, c);
            }
        }
        if (neg) {
            this->set_has_negative_cost();
        }
    }

    void add_transition_label(PerTransitionInfo<Updatable>* t, const int& label)
    {
        assert(label < (int) m_label_to_transitions.size());
        m_label_to_transitions[label].push_back(t);
    }

    void remove_transition_label(PerTransitionInfo<Updatable>* t, const int& label)
    {
        assert(label < (int) m_label_to_transitions.size());
        std::vector<PerTransitionInfo<Updatable>*>& ts = m_label_to_transitions[label];
        assert(std::count(ts.begin(), ts.end(), t));
        unsigned i = std::distance(ts.begin(), std::find(ts.begin(), ts.end(), t));
        if (i < ts.size() - 1) {
            ts[i] = ts.back();
        }
        ts.pop_back();
    }

protected:
    void update_labels_remove_unreachable_transitions()
    {
        for (int i = m_label_to_transitions.size() - 1; i >= 0; i--) {
            std::vector<PerTransitionInfo<Updatable>*>& ts = m_label_to_transitions[i];
            unsigned j = 0;
            for (unsigned k = 0; k < ts.size(); k++) {
                if (ts[k]->count >= 0) {
                    ts[j] = &this->m_transition_info[ts[k]->count];
                    j++;
                }
            }
            ts.resize(j);
        }
    }

    int get_label(const PerTransitionInfo<Updatable>* t) const
    {
        for (unsigned i = 0; i < m_label_to_transitions.size(); i++) {
            if (std::count(m_label_to_transitions[i].begin(), m_label_to_transitions[i].end(), t)) {
                return i;
            }
        }
        return -1;
    }

    std::vector<std::vector<PerTransitionInfo<Updatable>*> > m_label_to_transitions;
};





template<bool Updatable, bool Labeled>
__UpdatableHyperabstraction<Updatable, Labeled>::__UpdatableHyperabstraction(
        unsigned nodes,
        unsigned labels)
    : __LabeledHyperabstraction<Updatable, Labeled>(nodes, labels)
{
}

template<bool Updatable, bool Labeled>
void
__UpdatableHyperabstraction<Updatable, Labeled>::remember_transition_has_head(
        const Node& ,
        PerTransitionInfo<Updatable>* )
{
}

template<bool Updatable, bool Labeled>
void
__UpdatableHyperabstraction<Updatable, Labeled>::resize_head_relations()
{
}

template<bool Updatable, bool Labeled>
void
__UpdatableHyperabstraction<Updatable, Labeled>::update_head_relations_remove_unreachable_transitions()
{
}

template<bool Labeled>
class __UpdatableHyperabstraction<true, Labeled> : public __LabeledHyperabstraction<true, Labeled> {
public:
    __UpdatableHyperabstraction(unsigned nodes, unsigned labels)
        : __LabeledHyperabstraction<true, Labeled>(nodes, labels)
    {
        m_in_head.resize(nodes + 1);
    }

protected:
    void remember_transition_has_head(const Node& node,
                                      PerTransitionInfo<true>* t)
    {
        assert(node < m_in_head.size());
        m_in_head[node].push_back(t);
    }

    void resize_head_relations()
    {
        m_in_head.resize(this->num_nodes());
    }

    void update_head_relations_remove_unreachable_transitions()
    {
        assert(m_in_head.size() == this->num_nodes());
        for (int i = m_in_head.size() - 1; i >= 0; i--) {
            std::vector<PerTransitionInfo<true>*>& ts = m_in_head[i];
            unsigned j = 0;
            for (unsigned k = 0; k < ts.size(); k++) {
                if (ts[k]->count >= 0) {
                    ts[j] = &this->m_transition_info[ts[k]->count];
                    j++;
                }
            }
            ts.resize(j);
        }
    }

    SegmentedVector<std::vector<PerTransitionInfo<true> *> > m_in_head;
};




template<bool Updatable, bool Labeled>
__Hyperabstraction<Updatable, Labeled>::__Hyperabstraction(unsigned nodes,
                                                           unsigned labels)
    : __UpdatableHyperabstraction<Updatable, Labeled>(nodes, labels)
{
}

template<bool Updatable, bool Labeled>
void
__Hyperabstraction<Updatable, Labeled>::add_nodes(unsigned num_nodes_)
{
    this->m_in_tail.resize(this->num_nodes() + num_nodes_);
    this->resize_head_relations();
}

template<bool Updatable, bool Labeled>
void
__Hyperabstraction<Updatable, Labeled>::add_transition(
        const Node& head,
        const std::vector<Node>& tail,
        const int& label,
        const int& cost)
{
    if (cost < 0) {
        this->set_has_negative_cost();
    }
    this->m_transition_info.push_back(PerTransitionInfo<Updatable>(head, tail.empty() ? 1 : tail.size(), cost));
    PerTransitionInfo<Updatable>* t = &this->m_transition_info[this->m_transition_info.size() - 1];
    this->add_transition_label(t, label);
    this->remember_transition_has_head(head, t);
    if (tail.empty()) {
        this->m_in_tail[this->m_dummy].push_back(t);
    } else {
        for (int i = tail.size() - 1; i >= 0; i--) {
            assert(tail[i] < this->m_in_tail.size());
            this->m_in_tail[tail[i]].push_back(t);
        }
    }
    if (this->m_had_negative_cost) {
        this->m_transition_tails.insert(
                std::pair<const PerTransitionInfo<Updatable>*, std::vector<Node> >(
                    t, tail));
    }
}

template<bool Updatable, bool Labeled>
void
__Hyperabstraction<Updatable, Labeled>::compute_reachable(
        std::vector<bool>& reachable,
        const std::vector<Node>& from)
{
    assert(reachable.size() >= this->num_nodes());
    std::fill(reachable.begin(), reachable.end(), false);
    static std::vector<Node> queue; queue.reserve(this->num_nodes());
    assert(queue.empty());
    for (int i = from.size() - 1; i >= 0; i--) {
        assert(!reachable[from[i]]); // from shouldn't contain duplicates
        reachable[from[i]] = 1;
        queue.push_back(from[i]);
    }
    reachable[this->m_dummy] = 1;
    queue.push_back(this->m_dummy);
    for (int i = this->m_transition_info.size() - 1; i >= 0; i--) {
        this->m_transition_info[i].count = this->m_transition_info[i].tail_size;
    }
    while (!queue.empty()) {
        Node n = queue.back(); queue.pop_back();
        auto& ts = this->m_in_tail[n];
        for (int i = ts.size() - 1; i >= 0; i--) {
            auto& t = *ts[i];
            if (--t.count == 0 && !reachable[t.head]) {
                reachable[t.head] = true;
                queue.push_back(t.head);
            }
        }
    }
}

template<bool Updatable, bool Labeled>
bool
__Hyperabstraction<Updatable, Labeled>::propagate_reachability(
        std::vector<bool>& reachable,
        const std::vector<Node>& from,
        const std::vector<bool>& previous,
        const unsigned& shared)
{
    assert(reachable.size() >= this->num_nodes());
    assert(previous.size() >= this->num_nodes());
    std::fill(reachable.begin(), reachable.end(), false);
    static std::vector<Node> queue; queue.reserve(this->num_nodes());
    assert(queue.empty());
    for (int i = from.size() - 1; i >= 0; i--) {
        assert(!reachable[from[i]]); // from shouldn't contain duplicates
        reachable[from[i]] = previous[from[i]];
        if (reachable[from[i]]) {
            queue.push_back(from[i]);
        }
    }
    reachable[this->m_dummy] = 1;
    queue.push_back(this->m_dummy);
    for (int i = this->m_transition_info.size() - 1; i >= 0; i--) {
        this->m_transition_info[i].count = this->m_transition_info[i].tail_size;
    }
    while (!queue.empty()) {
        Node n = queue.back(); queue.pop_back();
        auto& ts = this->m_in_tail[n];
        for (int i = ts.size() - 1; i >= 0; i--) {
            auto& t = *ts[i];
            if (--t.count == 0 && !reachable[t.head] && previous[t.head]) {
                reachable[t.head] = true;
                queue.push_back(t.head);
            }
        }
    }
    return !std::equal(reachable.begin(), reachable.begin() + shared,
                       previous.begin(), previous.begin() + shared);
}

template<bool Updatable, bool Labeled>
void
__Hyperabstraction<Updatable, Labeled>::compute_distances(
        std::vector<int>& distances,
        const std::vector<Node>& from)
{
    static auto callback = [](const Node& ,
                              const PerTransitionInfo<Updatable>& ,
                              const int&) {
    };
    compute_distances(distances, from, callback);
// #ifndef NDEBUG
//     auto get_new_distance = [&](const Node& node) {
//         assert(node < this->m_in_head.size());
//         const auto& ts2 = this->m_in_head[node];
//         int val = INF;
//         for (int j = ts2.size() - 1; j >= 0; j--) {
//             const auto& t2 = *ts2[j];
//             if (t2.count == 0) {
//                 val = std::min(val, t2.value);
//             }
//         }
//         return val;
//     };
//     for (unsigned i = 0; i < this->num_nodes(); i++) {
//         assert(i == this->m_dummy || std::count(from.begin(), from.end(), i) || distances[i] == get_new_distance(i));
//     }
// #endif
}

template<bool Updatable, bool Labeled>
template<typename Callback>
void
__Hyperabstraction<Updatable, Labeled>::compute_distances(
        std::vector<int>& distances,
        const std::vector<Node>& from,
        Callback& callback)
{
    assert(distances.size() >= this->num_nodes());
    std::fill(distances.begin(), distances.end(), INF);
    static AdaptiveQueue<Node> queue;
    assert(queue.empty());
    for (int i = from.size() - 1; i >= 0; i--) {
        assert(distances[from[i]] == INF);
        distances[from[i]] = 0;
        queue.push(0, from[i]);
    }
    distances[this->m_dummy] = 0;
    queue.push(0, this->m_dummy);
    for (int i = this->m_transition_info.size() - 1; i >= 0; i--) {
        this->m_transition_info[i].count = this->m_transition_info[i].tail_size;
    }
    if (this->m_has_negative_cost) {
        compute_distances_negative_costs(distances, queue);
    } else {
        compute_distances(distances, queue, callback);
    }
}

template<bool Updatable, bool Labeled>
bool
__Hyperabstraction<Updatable, Labeled>::propagate_distances(
        std::vector<int>& distances,
        const std::vector<Node>& from,
        const std::vector<int>& previous_distances,
        const unsigned& shared)
{
    assert(distances.size() >= this->num_nodes());
    std::fill(distances.begin(), distances.end(), INF);
    static AdaptiveQueue<Node> queue;
    assert(queue.empty());
    for (int i = from.size() - 1; i >= 0; i--) {
        assert(distances[from[i]] == INF);
        distances[from[i]] = previous_distances[from[i]];
        if (distances[from[i]] != INF) {
            queue.push(previous_distances[from[i]], from[i]);
        }
    }
    distances[this->m_dummy] = 0;
    queue.push(0, this->m_dummy);
    for (int i = this->m_transition_info.size() - 1; i >= 0; i--) {
        this->m_transition_info[i].count = this->m_transition_info[i].tail_size;
    }
    return propagate_distances(distances, queue, previous_distances, shared);
}

template<bool Updatable, bool Labeled>
void
__Hyperabstraction<Updatable, Labeled>::compute_and_prune_unreachable(
        const std::vector<Node>& from)
{
    std::vector<int> distances(this->num_nodes());
    compute_distances(distances, from);
    prune_unreachable();
}

template<bool Updatable, bool Labeled>
unsigned
__Hyperabstraction<Updatable, Labeled>::prune_unreachable(
        const std::vector<bool>& reachable)
{
    assert(reachable.size() >= this->num_nodes());
    bool pruned = false;
    for (int i = this->num_transitions() - 1; i >= 0; i--) {
        auto& t = this->m_transition_info[i];
        t.count = 0;
        if (!reachable[t.head]) {
            t.count = 1;
            pruned = true;
        }
    }
    unsigned unreachable = 0;
    for (int node = this->num_nodes() - 1; node >= 0; node--) {
        if (reachable[node]) continue;
        unreachable++;
        auto& ts = this->m_in_tail[node];
        for (int i = ts.size() - 1; i >= 0; i--) {
            ts[i]->count = 1;
            pruned = true;
        }
    }
    if (pruned) {
        prune_unreachable();
    }
    return unreachable;
}

template<bool Updatable, bool Labeled>
void
__Hyperabstraction<Updatable, Labeled>::prune_unreachable()
{
    unsigned newidx = 0;
    for (unsigned oldidx = 0; oldidx < this->num_transitions(); oldidx++) {
        PerTransitionInfo<Updatable>& t = this->m_transition_info[oldidx];
        if (t.count == 0) {
            t.count = newidx++;
        } else {
            t.count = -1;
        }
    }

    if (newidx == this->m_transition_info.size()) {
        return;
    }

    this->update_labels_remove_unreachable_transitions();
    this->update_head_relations_remove_unreachable_transitions();

    for (int idx = this->m_in_tail.size() - 1; idx >= 0; idx--) {
        std::vector<PerTransitionInfo<Updatable>*>& ts = this->m_in_tail[idx];
        unsigned i = 0;
        for (unsigned j = 0; j < ts.size(); j++) {
            if (ts[j]->count >= 0) {
                ts[i] = &this->m_transition_info[ts[j]->count];
                i++;
            }
        }
        ts.resize(i);
    }

    newidx = 0;
    for (unsigned oldidx = 0; oldidx < this->num_transitions(); oldidx++) {
        const PerTransitionInfo<Updatable>& t = this->m_transition_info[oldidx];
        if (t.count >= 0) {
            if (newidx != oldidx) {
                this->m_transition_info[newidx] = this->m_transition_info[oldidx];
            }
            newidx++;
        }
    }

    this->m_transition_info.resize(newidx, PerTransitionInfo<Updatable>(-1, -1, -1));
}

template<bool Updatable, bool Labeled>
template<typename Callback>
void
__Hyperabstraction<Updatable, Labeled>::compute_distances(
        std::vector<int>& distances,
        AdaptiveQueue<Node>& queue,
        Callback& callback)
{
    while (!queue.empty()) {
        auto p = queue.pop();
        if (distances[p.second] < p.first) {
            continue;
        }
        const auto& transitions = this->m_in_tail[p.second];
        for (int i = transitions.size() - 1; i >= 0; i--) {
            auto& t = *transitions[i];
            if (--t.count == 0) {
                int value = p.first + t.cost;
                t.set_value(value);
                callback(p.second, t, value);
                if (distances[t.head] > value) {
                    distances[t.head] = value;
                    queue.push(value, t.head);
                }
            }
        }
    }
}

template<bool Updatable, bool Labeled>
void
__Hyperabstraction<Updatable, Labeled>::compute_distances_negative_costs(
        std::vector<int>& distances,
        AdaptiveQueue<Node>& queue)
{
    static std::vector<bool> unseen(this->num_nodes());;
    static std::priority_queue<std::pair<int, Node> > update_queue;
    std::fill(unseen.begin(), unseen.end(), true);
    assert(unseen.size() >= this->num_nodes());
    assert(update_queue.empty());
    while (!queue.empty()) {
        auto p = queue.pop();
        if (distances[p.second] < p.first) {
            continue;
        }
        unseen[p.second] = false;
        const auto& transitions = this->m_in_tail[p.second];
        for (int i = transitions.size() - 1; i >= 0; i--) {
            auto& t = *transitions[i];
            if (--t.count == 0) {
                int value = std::max(0, p.first + t.cost);
                t.set_value(value);
                if (value < distances[t.head]) {
                    distances[t.head] = value;
                    if (unseen[t.head]) {
                        queue.push(value, t.head);
                    } else {
                        update_queue.push(std::pair<int, Node>(value, t.head));
                    }
                }
            }
        }
    }
    propagate_decreased_distances(distances, update_queue);
}

template<bool Updatable, bool Labeled>
void
__Hyperabstraction<Updatable, Labeled>::propagate_decreased_distances(
        std::vector<int>& distances,
        std::priority_queue<std::pair<int, Node> >& queue)
{
    int value;
    int max_tail = 0;
    while (!queue.empty()) {
        auto p = queue.top(); queue.pop();
        if (distances[p.second] < p.first) {
            continue;
        }
        auto& in_tail_of = this->m_in_tail[p.second];
        for (int i = in_tail_of.size() - 1; i >= 0; i--) {
            PerTransitionInfo<Updatable>* t = in_tail_of[i];
            if (t->count) {
                continue;
            }
            max_tail = 0;
            const std::vector<Node>& tail = this->m_transition_tails[t];
            for (int j = tail.size() - 1; j >= 0; j--) {
                max_tail = std::max(max_tail, distances[tail[j]]);
            }
            assert(max_tail != INF);
            value = std::max(0, t->cost + max_tail);
            t->set_value(value);
            if (value < distances[t->head]) {
                distances[t->head] = value;
                queue.push(std::pair<int, Node>(value, t->head));
            }
        }
    }
}

template<bool Updatable, bool Labeled>
bool
__Hyperabstraction<Updatable, Labeled>::propagate_distances(
        std::vector<int>& distances,
        AdaptiveQueue<Node>& queue,
        const std::vector<int>& previous_distances,
        const unsigned& shared)
{
    while (!queue.empty()) {
        auto p = queue.pop();
        if (distances[p.second] < p.first) {
            continue;
        }

        int value = p.first;
        assert(value >= previous_distances[p.second]);

        const auto& transitions = this->m_in_tail[p.second];
        for (int i = transitions.size() - 1; i >= 0; i--) {
            auto& t = *transitions[i];
            int value_t = std::max(previous_distances[t.head], value);
            if (--t.count == 0) {
                t.set_value(value);
                if (value_t < distances[t.head]) {
                    distances[t.head] = value_t;
                    queue.push(value_t, t.head);
                }
            }
        }
    }

    return !std::equal(distances.begin(), distances.begin() + shared,
                       previous_distances.begin(), previous_distances.begin() + shared);
}

template<bool Updatable, bool Labeled>
void
__Hyperabstraction<Updatable, Labeled>::dump(std::ostream& out,
                                        bool swap_head_tail) const
{
    out << "digraph {" << std::endl;
    for (Node node = 0; node < this->num_nodes(); node++ ) {
        out << "  node" << node << " [shape=circle, label=\"n" << node << "\"];" << std::endl;
    }
    std::unordered_map<const PerTransitionInfo<Updatable>*, unsigned> ids;
    for (unsigned t = 0; t < this->num_transitions(); t++ ) {
        ids[&this->m_transition_info[t]] = t;
        out << "  transition" << t << " [shape=box, label=\"t" << t << ": "
            << this->get_label(&this->m_transition_info[t]) << "\"];" << std::endl;
    }
    for (Node node = 0; node < this->num_nodes(); node++) {
        for (const auto& t : this->m_in_tail[node]) {
            if (swap_head_tail) {
                out << "  transition" << ids[t] << " -> " << "node" << node << ";" << std::endl;
            } else {
                out << "  node" << node << " -> transition" << ids[t] << ";" << std::endl;
            }
        }
    }
    for (unsigned t = 0; t < this->num_transitions(); t++) {
        if (swap_head_tail) {
            out << "  node" << this->m_transition_info[t].head << " -> transition" << t << ";" << std::endl;
        } else {
            out << "  transition" << t << " -> " << "node" << this->m_transition_info[t].head << ";" << std::endl;
        }
    }
    out << "}" << std::endl;
}

}

#else
#include "hyperabstraction.h"
#include <functional>

namespace hyperabstraction {

template<>
void
__Hyperabstraction<true, false>::propagate(
        std::vector<int>& distances,
        std::priority_queue<std::pair<int, Node> >& queue_in,
        std::priority_queue<std::pair<int, Node> >& queue_out,
        const std::vector<bool>& from,
        const unsigned& actual_nodes)
{
#ifndef NDEBUG
    std::vector<int> inqueue(this->num_nodes(), 0);
    std::function<void(std::vector<int>&, std::priority_queue<std::pair<int, Node> >&)> collect;
    collect = [&collect](std::vector<int>& inqueue, std::priority_queue<std::pair<int, Node> >& queue) {
        if (queue.empty()) {
            return;
        }
        auto x = queue.top();
        queue.pop();
        if (inqueue[x.second] == 0) {
            inqueue[x.second] = x.first;
        }
        collect(inqueue, queue);
        queue.push(x);
    };
    collect(inqueue, queue_in);
    for (unsigned i = 0; i < m_in_head.size(); i++) {
        for (const auto& t : m_in_head[i]) {
            if (t->count == 0) {
                int val = 0;
                int val2 = 0;
                const auto& tail = m_transition_tails[t];
                for (const auto& j : tail) {
                    val = std::max(val, distances[j]);
                    val2 = std::max(val2, inqueue[j]);
                }
                assert(val <= t->value || val == val2);
            }
        }
    }
    // auto get_new_distance = [&](const Node& node) {
    //     assert(node < this->m_in_head.size());
    //     if (from[node]) return distances[node];
    //     const auto& ts2 = this->m_in_head[node];
    //     int val = INF;
    //     for (int j = ts2.size() - 1; j >= 0; j--) {
    //         const auto& t2 = *ts2[j];
    //         if (t2.count == 0) {
    //             val = std::min(val, t2.value);
    //         }
    //     }
    //     return val;
    // };
    // auto get = [&]() {
    //     int res = 0;
    //     for (unsigned node = 0; node < num_nodes(); node++) {
    //         res = std::max(res, get_new_distance(node));
    //     }
    //     return res;
    // };
    // std::cout << queue_in.top().first << " vs " << get() << std::endl;
    // assert(get() <= queue_in.top().first);
#endif
    assert(distances.size() >= this->num_nodes());
    while (!queue_in.empty()) {
        auto x = queue_in.top();
        queue_in.pop();
        if (distances[x.second] > x.first) {
            continue;
        }
        assert(queue_in.empty() || queue_in.top().first <= x.first);
        const auto& ts = this->m_in_tail[x.second];
        for (int i = ts.size() - 1; i >= 0; i--) {
            auto& t = *ts[i];
            bool changed = false;
            if (x.first == INF) {
                changed = t.count == 0;
                t.count++;
            } else if (t.count == 0 && t.value < x.first) {
                changed = true;
                t.value = x.first;
            }
            if (!from[t.head] && (changed || true)) {
                int val = t.value; 
                assert(t.head < this->m_in_head.size());
                const auto& ts2 = this->m_in_head[t.head];
                val = INF;
                for (int j = ts2.size() - 1; j >= 0; j--) {
                    const auto& t2 = *ts2[j];
                    if (t2.count == 0) {
                        val = std::min(val, t2.value);
                    }
                }
                if (val > distances[t.head]) {
                    assert(val <= x.first);
                    distances[t.head] = val;
                    queue_in.emplace(val, t.head);
                    if (t.head < actual_nodes) {
                        queue_out.emplace(val, t.head);
                    }
                }
            }
        }
    }
}



}

#endif
