#ifndef HYPERABSTRACTION_H
#define HYPERABSTRACTION_H

#include "types.h"
#include "../segmented_vector.h"
#include "../priority_queue.h"

#include <vector>
#include <queue>
#include <unordered_map>
#include <iostream>

namespace hyperabstraction {

template<bool Updatable>
struct __TransitionValue { void set_value(const int&) {} };

template<bool Updatable>
struct PerTransitionInfo : public __TransitionValue<Updatable> {
    int tail_size;
    int cost;
    Node head;

    int count;
    PerTransitionInfo(const Node& node,
                          const int& tail_size,
                          const int& cost);
    PerTransitionInfo(const PerTransitionInfo& t) = default;
    PerTransitionInfo(PerTransitionInfo&& t) = default;
    PerTransitionInfo& operator=(const PerTransitionInfo& t) = default;
    PerTransitionInfo& operator=(PerTransitionInfo&& t) = default;
};


template<bool Updatable>
class __HyperabstractionBase {
public:
    __HyperabstractionBase(unsigned nodes,
                           unsigned labels);

    void set_transition_cost(PerTransitionInfo<Updatable>* t, const int& cost);

    const PerTransitionInfo<Updatable>& get_transition(unsigned id) const;
    PerTransitionInfo<Updatable>& get_transition(unsigned id);
    const std::vector<PerTransitionInfo<Updatable> >& get_transitions_with_tail(const Node& node) const;

    unsigned num_nodes() const;
    unsigned num_transitions() const;
    unsigned num_hypertransitions_strict() const;
    unsigned num_labels() const;

    bool has_negative_cost() const;
    void set_has_negative_cost();
protected:
    SegmentedVector<PerTransitionInfo<Updatable> > m_transition_info;
    SegmentedVector<std::vector<PerTransitionInfo<Updatable>* > > m_in_tail;
    std::unordered_map<const PerTransitionInfo<Updatable>*, std::vector<Node> > m_transition_tails;
    Node m_dummy;
    unsigned m_num_labels;
    bool m_has_negative_cost;
    bool m_had_negative_cost;
};

template<bool Updatable, bool Labeled>
class __LabeledHyperabstraction : public __HyperabstractionBase<Updatable> {
public:
    __LabeledHyperabstraction(unsigned nodes, unsigned labels);
    void set_costs(const std::vector<int>& costs);
    void add_transition_label(PerTransitionInfo<Updatable>* t, const int& l);
    void remove_transition_label(PerTransitionInfo<Updatable>* t, const int& l);
protected:
    void update_labels_remove_unreachable_transitions();
    int get_label(const PerTransitionInfo<Updatable>* t) const;
};

template<bool Updatable, bool Labeled>
class __UpdatableHyperabstraction : public __LabeledHyperabstraction<Updatable, Labeled> {
public:
    __UpdatableHyperabstraction(unsigned nodes, unsigned labels);
protected:
    void remember_transition_has_head(const Node& head, PerTransitionInfo<Updatable>* t);
    void resize_head_relations();
    void update_head_relations_remove_unreachable_transitions();
};

template<bool Updatable, bool Labeled>
class __Hyperabstraction : public __UpdatableHyperabstraction<Updatable, Labeled> {
public:
    __Hyperabstraction(unsigned num_nodes, unsigned num_labels);

    void add_nodes(unsigned num_nodes);
    void add_transition(const Node& head,
                        const std::vector<Node>& tail,
                        const int& label,
                        const int& cost);

    void compute_reachable(std::vector<bool>& reachable,
                              const std::vector<Node>& from);
    bool propagate_reachability(std::vector<bool>& reachable,
                                const std::vector<Node>& from,
                                const std::vector<bool>& previous,
                                const unsigned& shared);
    
    void compute_distances(std::vector<int>& distances,
                           const std::vector<Node>& from);
    template<typename Callback>
    void compute_distances(std::vector<int>& distances,
                          const std::vector<Node>& from,
                          Callback& callback);

    bool propagate_distances(std::vector<int>& distances,
                             const std::vector<Node>& from,
                             const std::vector<int>& previous_distances,
                             const unsigned& shared);

    void propagate(std::vector<int>& distances,
                   std::priority_queue<std::pair<int, Node> >& queue_in,
                   std::priority_queue<std::pair<int, Node> >& queue_out,
                   const std::vector<bool>& from,
                   const unsigned& actual_nodes);

    void compute_and_prune_unreachable(const std::vector<Node>& from);
    unsigned prune_unreachable(const std::vector<bool>& reachable);
    void prune_unreachable();

    void dump(std::ostream& out, bool swap_head_tail = false) const;
protected:
    template<typename Callback>
    void compute_distances(std::vector<int>& distances,
                           AdaptiveQueue<Node>& queue,
                           Callback& callback);
    void compute_distances_negative_costs(std::vector<int>& distances,
                                          AdaptiveQueue<Node>& queue);
    void propagate_decreased_distances(std::vector<int>& distances,
                                       std::priority_queue<std::pair<int, Node> >& queue);
    bool propagate_distances(std::vector<int>& distances,
                             AdaptiveQueue<Node>& queue,
                             const std::vector<int>& previous_distances,
                             const unsigned& shared);
};

using Hyperabstraction = __Hyperabstraction<false, false>;
using LabeledHyperabstraction = __Hyperabstraction<false, true>;
using UpdatableHyperabstraction = __Hyperabstraction<true, false>;

}

#include "hyperabstraction.cc"

#endif
