#include "hyperabstraction_generator_statistics.h"
#include "generic_hyperabstraction_generator.h"
#include <limits>

namespace hyperabstraction {

HyperabstractionGeneratorStatistics::HyperabstractionGeneratorStatistics()
    : m_num_nodes(0)
    , m_num_transitions(0)
    , m_num_hypertransitions(0)
    , m_num_tau_transitions(0)
    , m_num_tau_nodes(0)
    , m_num_dinf_nodes(0)
    , m_num_dinf_transitions(0)
    , m_num_pruned_transitions(0)
    , m_transition_pruning(TransitionPruningMethod::DISABLED)
    , m_cache_tau_nodes(false)
    , m_t_tau_max_nodes(0)
    , m_g_tau_max_nodes(0)
    , m_prune_inf(false)
{
    m_construction_timer.reset();
    m_construction_timer.stop();
    m_postprocessing_timer.reset();
    m_postprocessing_timer.stop();
    m_num_t0_transitions = 0;
    m_num_t1_transitions = 0;
    m_num_t2_transitions = 0;
    m_num_t3_transitions = 0;
    m_num_t4_transitions = 0;
    m_num_lt4_transitions = 0;
}
    
unsigned
HyperabstractionGeneratorStatistics::num_tau_transitions() const
{
    return m_num_tau_transitions;
}

unsigned
HyperabstractionGeneratorStatistics::num_tau_nodes() const
{
    return m_num_tau_nodes;
}

double
HyperabstractionGeneratorStatistics::construction_time() const
{
    return m_construction_timer();
}

void
HyperabstractionGeneratorStatistics::set_transition_pruning_method(const TransitionPruningMethod& method)
{
    m_transition_pruning = method;
}

void
HyperabstractionGeneratorStatistics::set_caching_tau_nodes(bool cache)
{
    m_cache_tau_nodes = cache;
}

void
HyperabstractionGeneratorStatistics::set_tau_max_nodes(const unsigned& t,
                                                       const unsigned& g)
{
    m_t_tau_max_nodes = t;
    m_g_tau_max_nodes = g;
}

void
HyperabstractionGeneratorStatistics::set_prune_inf(bool prune)
{
    m_prune_inf = prune;
}

void
HyperabstractionGeneratorStatistics::print(bool forward, std::ostream& out) const
{
    out << (forward ? "Forward" : "Backward") << " hyperabstraction statistics {" << std::endl;
    out << "  Transition pruning method: "
        << (m_transition_pruning == TransitionPruningMethod::DISABLED ? "disabled" : "")
        << (m_transition_pruning == TransitionPruningMethod::DUPLICATES ? "duplicates" : "")
        << (m_transition_pruning == TransitionPruningMethod::SUBSUMED ? "subsumed" : "")
        << std::endl;
    out << "  Caching tau nodes: "
        << (m_cache_tau_nodes ? "true" : "false") << std::endl;
    out << "  Max nodes represented per tau node for transitions: ";
    if (m_t_tau_max_nodes >= (unsigned) std::numeric_limits<int>::max()) {
        out << "infinity" << std::endl;
    } else {
        out << m_t_tau_max_nodes << std::endl;
    }
    out << "  Max nodes represented per tau node for goal: ";
    if (m_g_tau_max_nodes >= (unsigned) std::numeric_limits<int>::max()) {
        out << "infinity" << std::endl;
    } else {
        out << m_g_tau_max_nodes << std::endl;
    }
    // out << "  Precompute and prune if distance = infinity: "
    //     << (m_prune_inf ? "true" : "false")
    //     << std::endl;
    out << "  Nodes: " << m_num_nodes << std::endl;
    out << "  Transitions: " << m_num_transitions << std::endl;
    out << "  Hypertransitions (strict): " << m_num_hypertransitions << std::endl;
    out << "  Transitions with tail size "
        << "0: "  << m_num_t0_transitions << ", "
        << "1: "  << m_num_t1_transitions << ", "
        << "2: "  << m_num_t2_transitions << ", "
        << "3: "  << m_num_t3_transitions << ", "
        << "4: "  << m_num_t4_transitions << ", "
        << ">4: " << m_num_lt4_transitions << std::endl;
    out << "  Tau nodes: " << m_num_tau_nodes << std::endl;
    out << "  Tau transitions: " << m_num_tau_transitions << std::endl;;
    // out << "  D=inf nodes: " << m_num_dinf_nodes << std::endl;
    // out << "  D=inf transitions: " << m_num_dinf_transitions << std::endl;
    out << "  Transitions pruned during construction: " << m_num_pruned_transitions << std::endl;
    out << "  Postprocessing time: " << m_postprocessing_timer << std::endl;
    out << "  Construction time: " << m_construction_timer << std::endl;
    out << "}" << std::endl;
}

void
HyperabstractionGeneratorStatistics::begin_construction()
{
    m_construction_timer.reset();
    m_construction_timer.resume();
    m_num_nodes = 0;
    m_num_transitions = 0;
    m_num_hypertransitions = 0;
    m_num_tau_nodes = 0;
    m_num_tau_transitions = 0;
}

void
HyperabstractionGeneratorStatistics::end_construction(unsigned nodes,
                                                      unsigned transitions,
                                                      unsigned hypertransitions)
{
    m_construction_timer.stop();
    m_num_nodes = nodes;
    m_num_transitions = transitions;
    m_num_hypertransitions = hypertransitions;
}

void
HyperabstractionGeneratorStatistics::begin_postprocessing()
{
    m_postprocessing_timer.reset();
    m_postprocessing_timer.resume();
}

void
HyperabstractionGeneratorStatistics::end_postprocessing()
{
    m_postprocessing_timer.stop();
}

void
HyperabstractionGeneratorStatistics::inc_tau_transitions(unsigned num)
{
    m_num_tau_transitions += num;
}

void
HyperabstractionGeneratorStatistics::inc_tau_nodes(unsigned num)
{
    m_num_tau_nodes += num;
}

void
HyperabstractionGeneratorStatistics::inc_dinf_nodes(unsigned num)
{
    m_num_dinf_nodes += num;
}

void
HyperabstractionGeneratorStatistics::inc_dinf_transitions(unsigned num)
{
    m_num_dinf_transitions += num;
}

void
HyperabstractionGeneratorStatistics::inc_pruned_transitions(unsigned num)
{
    m_num_pruned_transitions += num;
}

void
HyperabstractionGeneratorStatistics::inc_transitions(const unsigned& tail_size)
{
    if (tail_size == 0) {
        m_num_t0_transitions++;
    } else if (tail_size == 1) {
        m_num_t1_transitions++;
    } else if (tail_size == 2) {
        m_num_t2_transitions++;
    } else if (tail_size == 3) {
        m_num_t3_transitions++;
    } else if (tail_size == 4) {
        m_num_t4_transitions++;
    } else {
        m_num_lt4_transitions++;
    }
}

}
