#ifdef HYPERABSTRACTION_TRANSITION_GENERATOR_H

#include <algorithm>
#include <cassert>

namespace hyperabstraction {

template<typename Abstraction>
TransitionGeneratorNoPrune<Abstraction>::TransitionGeneratorNoPrune(
        Abstraction& a,
        HyperabstractionGeneratorStatistics& stats)
    : abstraction(a)
    , stats(stats)
{}

template<typename Abstraction>
void
TransitionGeneratorNoPrune<Abstraction>::operator()(
        const Node& node,
        const std::vector<Node>& tail,
        const int& label,
        const int& cost)
{
    abstraction.add_transition(node, tail, label, cost);
    stats.inc_transitions(tail.size());
}

template<typename Abstraction>
TransitionGeneratorPruneDuplicate<Abstraction>::TransitionGeneratorPruneDuplicate(
        Abstraction& a,
        HyperabstractionGeneratorStatistics& stat)
    : abstraction(a)
    , transition_set(1e5, TransitionGeneratorPruneDuplicate<Abstraction>::Hasher(transitions), TransitionGeneratorPruneDuplicate<Abstraction>::Comparator(transitions))
    , stats(stat)
{}

template<typename Abstraction>
void
TransitionGeneratorPruneDuplicate<Abstraction>::operator()(
        const Node& head,
        const std::vector<Node>& tail,
        const int& label,
        const int& cost)
{
    transitions.push_back(tail);
    auto& identifier = transitions[transitions.size() - 1];
    std::sort(identifier.begin(), identifier.end());
    identifier.push_back(head);
    auto insrt = transition_set.insert(transitions.size() - 1);
    if (insrt.second) {
        // if (true || head < 401) {
        //     std::cout << "(" << head << " -> [";
        //     for (unsigned i = 0; i < identifier.size() - 1; i++) {
        //         //if (identifier[i] >= 401) break;
        //         std::cout << (i > 0 ? ", " : "") << identifier[i];
        //     }
        //     std::cout << "])" << std::endl;
        // }
        // std::cout << "[";
        // for (unsigned i = 0; i < identifier.size(); i++) {
        //     std::cout << (i > 0 ? ", " : "") << identifier[i];
        // }
        // std::cout << "] -> " << abstraction.num_transitions() << std::endl;
        transition_ids.push_back(abstraction.num_transitions());
        abstraction.add_transition(head, tail, label, cost);
        stats.inc_transitions(tail.size());
        return;
    }
    stats.inc_pruned_transitions();
    auto& t = abstraction.get_transition(transition_ids[*insrt.first]);
    abstraction.add_transition_label(&t, label);
    abstraction.set_transition_cost(&t, std::min(t.cost, cost));
    transitions.resize(transitions.size() - 1);
}

template<typename Abstraction, bool Forward>
TransitionGeneratorPruneSubsumed<Abstraction, Forward>::TransitionGeneratorPruneSubsumed(
        Abstraction& a,
        HyperabstractionGeneratorStatistics& stat)
    : abstraction(a), stats(stat)
{
    transitions.resize(a.num_nodes());
}

template<typename Abstraction, bool Forward>
void
TransitionGeneratorPruneSubsumed<Abstraction, Forward>::operator()(
        const Node& head,
        const std::vector<Node>& tail,
        const int& label,
        const int& cost)
{
    static std::vector<Node> sorted;
    assert(sorted.empty());
    sorted.insert(sorted.end(), tail.begin(), tail.end());
    std::sort(sorted.begin(), sorted.end());
    if (head >= transitions.size()) {
        transitions.resize(abstraction.num_nodes());
    }
    bool is_subsumed = false;
    std::map<int, utils::UBTreeMap<Node, int>* >& ts = transitions[head];
    auto it = ts.begin();
    for (; it != ts.end() && it->first <= cost; it++) {
        if ((Forward && it->second->contains_superset(sorted))
            || (!Forward && it->second->contains_subset(sorted))) {
            is_subsumed = true;
            break;
        }
    }
    if (!is_subsumed) {
        for (; it != ts.end(); it++) {
            auto mark = [this](const int& idx) {
                abstraction.get_transition(idx).count = 1;
            };
            if (Forward) {
                it->second->for_every_subset_do(sorted, mark);
            } else {
                it->second->for_every_superset_do(sorted, mark);
            }
        }
        auto insrt = ts.insert(std::pair<int, utils::UBTreeMap<Node, int>*>(cost, NULL));
        if (insrt.second) {
            _local_transitions.emplace_back();
            insrt.first->second = &_local_transitions.back();
        }
        insrt.first->second->insert(sorted, abstraction.num_transitions());
        abstraction.add_transition(head, tail, label, cost);
    }
    sorted.clear();
}

template<typename Abstraction, bool Forward>
void
TransitionGeneratorPruneSubsumed<Abstraction, Forward>::finalize()
{
    abstraction.prune_unreachable();
}

}

#endif
