#ifndef PDB_HYPERABSTRACTION_GENERATOR_H
#define PDB_HYPERABSTRACTION_GENERATOR_H

#include "hyperabstraction.h"
#include "hyperabstraction_generator.h"
#include "generic_hyperabstraction_generator.h"
#include "../pdbs/types.h"

#include "../cegar/domains.h"

#include <vector>

namespace hyperabstraction {

class PDBNodeMapper {
public:
    PDBNodeMapper(const std::shared_ptr<AbstractTask> task,
                  const pdbs::Pattern& pattern,
                  unsigned offset,
                  bool enumerate_tail = false);
    template<typename Callback>
    void for_each_intersecting_node(std::vector<int>& head,
                                    std::vector<int>& tail,
                                    Callback& callback) const;
    bool get_compatible_nodes(std::vector<Node>& result,
                                const std::vector<int>& partial_state,
                                unsigned max_nodes) const;
    bool get_compatible_nodes(std::vector<Node>& result,
                                const cegar::Domains& partial_state,
                                unsigned max_nodes) const;
    std::vector<int> compute_intersection(Node node,
                                            const std::vector<int>& partial_state) const;
    cegar::Domains compute_intersection(Node node,
                                        const cegar::Domains& cegar_state) const;
    Node get_node(const std::vector<int>& partial_state) const;
    Node get_node(const GlobalState& state) const;
    unsigned get_num_compatible_nodes(const std::vector<int>& partial_state,
                                      unsigned max_nodes = (unsigned) -1) const;
    unsigned get_num_compatible_nodes(const cegar::Domains& cartesian_state,
                                      unsigned max_nodes = (unsigned) -1) const;
    unsigned num_nodes() const;
private:
    template<bool Enumerative, typename Callback>
    void for_each_intersecting_node(std::vector<int>& head,
                                    std::vector<int>& tail,
                                    Callback& callback,
                                    int i,
                                    Node base) const;
    template<typename Callback>
    void for_each_compatible_node(const std::vector<int>& partial_state,
                                    const Callback& callback,
                                    int i,
                                    Node base) const;
    template<typename Callback>
    void for_each_compatible_node(const cegar::Domains& partial_state,
                                    const Callback& callback,
                                    int i,
                                    Node base) const;

    const bool c_enumerate_tail;
    const std::shared_ptr<AbstractTask> m_task;
    pdbs::Pattern m_pattern;
    std::vector<unsigned> m_sizes;
    unsigned m_offset;
    unsigned m_num_nodes;
};

class PDBHyperabstractionGenerator : public HyperabstractionGenerator {
public:
    PDBHyperabstractionGenerator(const options::Options& opts);

    virtual HyperabstractionGeneratorResult<Hyperabstraction> generate_backward() override;
    virtual HyperabstractionGeneratorResult<LabeledHyperabstraction> generate_backward_labeled() override;
    virtual HyperabstractionGeneratorResult<UpdatableHyperabstraction> generate_backward_updatable() override;
    virtual HyperabstractionGeneratorResult<Hyperabstraction> generate_forward() override;
    virtual HyperabstractionGeneratorResult<LabeledHyperabstraction> generate_forward_labeled() override;
    virtual HyperabstractionGeneratorResult<UpdatableHyperabstraction> generate_forward_updatable() override;

    virtual void get_nodes(std::vector<Node>& nodes, const GlobalState& state) override;
    virtual unsigned num_nodes() const override;

    static void add_options_to_parser(options::OptionParser& parser);

private:
    template<typename Abstraction, bool Direction>
    HyperabstractionGeneratorResult<Abstraction> generate();

    std::vector<PDBNodeMapper> m_node_mappers;
    unsigned m_num_nodes;
};


template<typename Abstraction, bool Direction>
HyperabstractionGeneratorResult<Abstraction>
PDBHyperabstractionGenerator::generate()
{
    return generate_hyperabstraction<Abstraction>(
            m_task,
            Direction,
            c_tau_max_nodes,
            c_tau_max_nodes_goal,
            c_cache_tau_nodes,
            c_transition_pruning,
            false,
            c_transition_enumeration_depth,
            c_goal_enumeration_depth,
            c_enumeration_max_nodes,
            m_node_mappers);
}

}





namespace hyperabstraction {


template<typename Callback>
void
PDBNodeMapper::for_each_intersecting_node(
        std::vector<int>& head,
        std::vector<int>& tail,
        Callback& callback) const
{
    bool not_interesting = true;
    for (int i = m_pattern.size() - 1; i >= 0; i--) {
        int var = m_pattern[i];
        if ((head[var] != -1 && head[var] != tail[var])
                || (head[var] == -1 && tail[var] != -1)) {
            not_interesting = false;
            break;
        }
    }
    if (not_interesting) {
        return;
    }
    if (c_enumerate_tail) {
        for_each_intersecting_node<true>(head, tail, callback, m_pattern.size() - 1, m_offset);
    } else {
        for_each_intersecting_node<false>(head, tail, callback, m_pattern.size() - 1, m_offset);
    }
}

template<bool Enumerative, typename Callback>
void
PDBNodeMapper::for_each_intersecting_node(
        std::vector<int>& head,
        std::vector<int>& tail,
        Callback& callback,
        int i,
        Node base) const
{
    while (i >= 0 && head[m_pattern[i]] != -1 && (!Enumerative || tail[m_pattern[i]] != -1)) {
        base += (head[m_pattern[i]] * m_sizes[i]);
        i--;
    }
    if (i < 0) {
// #ifndef NDEBUG
//         auto dumpv = [](const auto& v) {
//             std::cout << "[";
//             bool sep = false;
//             for (unsigned i = 0; i < v.size(); i++) {
//                 // if ((unsigned) v[i] == (unsigned) -1) continue;
//                 if (sep) std::cout << ", ";
//                 if (v.size() > 5)  std::cout << i << ":";
//                 std::cout << v[i];
//                 sep = true;
//             }
//             std::cout << "]" << std::endl;
//         };
//         std::cout << "pattern = ";
//         dumpv(m_pattern);
//         std::cout << "sizes = ";
//         dumpv(m_sizes);
//         std::cout << "head = ";
//         dumpv(head);
//         std::cout << "tail = ";
//         dumpv(tail);
//         std::cout << "head_node = " << (get_node(head) - m_offset) << " (+" << m_offset << ")" << std::endl;
//         std::cout << "node = " << (base - m_offset) << std::endl;
//         std::vector<int> thisstate;
//         Node n = base - m_offset;
//         for (int j = m_pattern.size() - 1; j >= 0; j--) {
//             int var = m_pattern[j];
//             int val = n / m_sizes[j];
//             n -= (val * m_sizes[j]);
//             thisstate.push_back(val);
//             if (head[var] != -1 && head[var] != val) {
//                 std::cout << "mistmatch at var " << var
//                           << ": val should be " << tail[var]
//                           << " but is " << val << std::endl;
//             }
//             assert(head[var] == -1 || head[var] == val);
//         }
//         std::reverse(thisstate.begin(), thisstate.end());
//         std::cout << "node#" << (base - m_offset) << " = ";
//         dumpv(thisstate);
// #endif
        callback(base, tail);
    } else {
        int var = m_pattern[i];
        int hval = head[var];
        bool head_defined = hval != -1;
        bool tail_defined = tail[var] != -1;
        for (int val = m_task->get_variable_domain_size(var) - 1; val >= 0; val--) {
            if (!head_defined) {
                head[var] = val;
                hval = val;
                if (!tail_defined) {
                    tail[var] = val;
                }
            } else if (Enumerative && !tail_defined) {
                tail[var] = val;
            }
            for_each_intersecting_node<Enumerative>(head, tail, callback, i - 1, base + (hval * m_sizes[i]));
        }
        if (!head_defined) {
            head[var] = -1;
            if (!tail_defined) {
                tail[var] = -1;
            }
        } else if (Enumerative && !tail_defined) {
            tail[var] = -1;
        }
    }
}

template<typename Callback>
void
PDBNodeMapper::for_each_compatible_node(
        const std::vector<int>& partial_state,
        const Callback& callback,
        int i,
        Node base) const
{
    while (i >= 0 && partial_state[m_pattern[i]] != -1) {
        base += (partial_state[m_pattern[i]] * m_sizes[i]);
        i--;
    }
    if (i < 0) {
        callback(base);
    } else {
        for (int val = m_task->get_variable_domain_size(m_pattern[i]) - 1; val >= 0; val--) {
            for_each_compatible_node(partial_state, callback, i - 1, base + (val * m_sizes[i]));
        }
    }
}

template<typename Callback>
void
PDBNodeMapper::for_each_compatible_node(
        const cegar::Domains& partial_state,
        const Callback& callback,
        int i,
        Node base) const
{
    assert(i >= 0);
    const int& var = m_pattern[i];
    const int& size = m_sizes[i];
    if (i == 0) {
        for (int val = m_task->get_variable_domain_size(var) - 1; val >= 0; val--) {
            if (partial_state.test(var, val)) {
                callback(base + val * size);
            }
        }
    } else {
        for (int val = m_task->get_variable_domain_size(var) - 1; val >= 0; val--) {
            if (partial_state.test(var, val)) {
                for_each_compatible_node(partial_state, callback, i - 1, base + val * size);
            }
        }
    }
}

}

#endif
