#include "saturated_cost_partitioning_finder.h"

#include "../cegar/cost_saturation.h"
#include "../pdbs/pattern_generator.h"

#include "../cost_saturation/types.h"
#include "../cost_saturation/cost_partitioning_collection_generator.h"
#include "../cost_saturation/cost_partitioning_heuristic.h"
#include "../cost_saturation/abstraction.h"
#include "../cost_saturation/cartesian_abstraction_generator.h"
#include "../cost_saturation/utils.h"
#include "../cost_saturation/cost_partitioned_heuristic.h"
#include "../cost_saturation/projection.h"

#include "../utils/memory.h"
#include "../utils/system.h"
#include "../task_utils/task_properties.h"

#include "../option_parser.h"
#include "../plugin.h"

#include <memory>

namespace hyperabstraction {

SaturatedCostPartitioningFinder::SaturatedCostPartitioningFinder(
        const options::Options& opts)
    : CostPartitioningFinder(opts)
{
#ifndef NDEBUG
    m_partitions_extracted = false;
#endif
    cost_saturation::CostPartitioningCollectionGenerator generator(
            cost_saturation::get_cp_collection_generator_from_options(opts));

    TaskProxy task_proxy(*m_task);

    std::vector<std::unique_ptr<cost_saturation::Abstraction> > abstractions;

    if (opts.contains("cartesian")) {
        std::shared_ptr<cegar::CostSaturation> cartesian =
            opts.get<std::shared_ptr<cegar::CostSaturation>>("cartesian");
        std::vector<std::unique_ptr<cost_saturation::Abstraction> > abstr =
            cost_saturation::generate_abstractions(*cartesian, task_proxy);
        for (unsigned i = 0; i < abstr.size(); i++) {
            abstractions.push_back(std::move(abstr[i]));
        }
    }

    if (opts.contains("patterns")) {
        std::shared_ptr<pdbs::PatternCollectionGenerator> gen =
            opts.get<std::shared_ptr<pdbs::PatternCollectionGenerator> >("patterns");
        pdbs::PatternCollectionInformation pattern_collection_info =
            gen->generate(m_task);
        std::shared_ptr<pdbs::PatternCollection> patterns =
            pattern_collection_info.get_patterns();
        for (const pdbs::Pattern &pattern : *patterns) {
            abstractions.push_back(
                utils::make_unique_ptr<cost_saturation::Projection>(task_proxy, pattern));
        }
    }

    if (!abstractions.empty()) {
        std::vector<CostPartitioning> partitions;

        cost_saturation::CPCollectionCallback callback = [&](
                unsigned mode,
                bool before,
                const std::vector<int>& costs) {
            if (mode == cost_saturation::CallbackCaller::POP_BACK) {
                partitions.pop_back();
                return;
            } else if (mode ==  cost_saturation::CallbackCaller::POP_PREVIOUS) {
                assert(partitions.size() >= 2);
                partitions[partitions.size() - 2].swap(partitions.back());
                partitions.pop_back();
                return;
            }
            if (mode == cost_saturation::CallbackCaller::NEW) {
                partitions.emplace_back();
            }
            if (before) {
                partitions.back().push_back(costs);
                partitions.back().back().push_back(0);
            } else {
                std::vector<int>& rem = partitions.back().back();
                for (int i = costs.size() - 1; i >= 0; i--) {
                    rem[i] -= costs[i];
                }
            }
        };

        std::vector<int> costs = task_properties::get_operator_costs(task_proxy);
        generator.get_cost_partitionings(
                task_proxy,
                abstractions,
                costs,
                cost_saturation::compute_saturated_cost_partitioning,
                callback);

        m_partitions = std::move(partitions);
    } else {
        std::vector<int> costs = task_properties::get_operator_costs(task_proxy);
        costs.push_back(0);

        m_partitions.m_cost_partitions.emplace_back(std::move(costs));
        m_partitions.m_collection.emplace_back(std::vector<unsigned>{0});
        //std::cerr << "No abstractions for cost partitioning computation given!" << std::endl;
        //utils::exit_with(utils::ExitCode::CRITICAL_ERROR);
    }
}

CostPartitioningCollection
SaturatedCostPartitioningFinder::compute_cost_partitioning(
    LabeledHyperabstraction& ,
    const std::vector<Node>& ,
    const std::vector<Node>& )
{
#ifndef NDEBUG
    assert(!m_partitions_extracted);
    m_partitions_extracted = true;
#endif
    return std::move(m_partitions);
}

void
SaturatedCostPartitioningFinder::add_options_to_parser(options::OptionParser& parser)
{
    parser.add_option<std::shared_ptr<cegar::CostSaturation> >("cartesian",
            "",
            options::OptionParser::NONE);
    parser.add_option<std::shared_ptr<pdbs::PatternCollectionGenerator> >(
            "patterns",
            "",
            options::OptionParser::NONE);
            // "systematic(pattern_max_size=1, only_interesting_patterns=true)");
    cost_saturation::add_cost_partitioning_collection_options_to_parser(parser);
    parser.add_option<std::shared_ptr<cost_saturation::CostPartitioningGenerator>>(
        "orders",
        "cost partitioning generator",
        "greedy(scoring_function=max_heuristic_per_stolen_costs, dynamic=false, random_seed=0)");
    CostPartitioningFinder::add_options_to_parser(parser);
}

}

static PluginShared<hyperabstraction::CostPartitioningFinder> _plugin(
        "saturated_cost_partitioning",
        options::parse_shared<hyperabstraction::CostPartitioningFinder,
                              hyperabstraction::SaturatedCostPartitioningFinder>);
