#ifdef HYPERABSTRACTION_UTILS_H

namespace hyperabstraction {

template<typename H1, typename H2>
int
compute_bidirectional_distances(
        std::vector<int>& distances2,
        const std::vector<Node>& fw_init,
        const std::vector<Node>& bw_goal,
        const std::vector<Node>& state_nodes,
        const unsigned& num_shared_nodes,
        H1& fw_abstraction,
        H2& bw_abstraction)
{
    assert(distances2.size() == std::max(fw_abstraction.num_nodes(), bw_abstraction.num_nodes()));
    static std::vector<int> distances1(distances2.size());
    if (get_max_distance(distances2, state_nodes) == INF) {
        return INF;
    }
    bw_abstraction.compute_distances(distances1, state_nodes);
    if (get_max_distance(distances1, state_nodes) == INF) {
        return INF;
    }
    for (int i = num_shared_nodes - 1; i >= 0; i--) {
        if (distances1[i] == INF || distances2[i] == INF) {
            distances2[i] = INF;
        } else {
            distances2[i] += distances1[i];
        }
    }
    bool updated_fw = true;
    bool updated_bw = true;
    bool fw = true;
    int res = -1;
    while (res != INF && (updated_fw || updated_bw)) {
        if ((!fw && !updated_bw) || (fw && !updated_fw)) {
            fw = !fw;
        }
        if (fw) {
            std::fill(distances2.begin() + num_shared_nodes, distances2.end(), 0);
            updated_fw = fw_abstraction.propagate_distances(
                    distances1,
                    fw_init,
                    distances2,
                    num_shared_nodes);
            res = get_max_distance(distances1, state_nodes);
        } else {
            std::fill(distances2.begin() + num_shared_nodes, distances2.end(), 0);
            updated_bw = bw_abstraction.propagate_distances(
                    distances1,
                    state_nodes,
                    distances2,
                    num_shared_nodes);
            res = get_max_distance(distances1, bw_goal);
        }
        fw = !fw;
        distances2.swap(distances1);
    }
    return res;
}

template<typename H1, typename H2>
bool
compute_bidirectional_reachability(
        std::vector<bool>& reachable2,
        const std::vector<Node>& fw_init,
        const std::vector<Node>& bw_goal,
        const std::vector<Node>& state_nodes,
        const unsigned& num_shared_nodes,
        H1& fw_abstraction,
        H2& bw_abstraction)
{
    assert(reachable2.size() == std::max(fw_abstraction.num_nodes(), bw_abstraction.num_nodes()));
    static std::vector<bool> reachable1(reachable2.size());
    if (!is_reachable(reachable2, state_nodes)) {
        return false;
    }
    bw_abstraction.compute_reachable(reachable1, state_nodes);
    if (!is_reachable(reachable1, bw_goal)) {
        return false;
    }
    for (int i = num_shared_nodes - 1; i >= 0; i--) {
        reachable2[i] = reachable1[i] && reachable2[i];;
    }
    bool updated_fw = true;
    bool updated_bw = true;
    bool fw = true;
    bool res = true;
    while (res && (updated_fw || updated_bw)) {
        if ((!fw && !updated_bw) || (fw && !updated_fw)) {
            fw = !fw;
        }
        if (fw) {
            std::fill(reachable2.begin() + num_shared_nodes, reachable2.end(), true);
            updated_fw = fw_abstraction.propagate_reachability(
                    reachable1,
                    fw_init,
                    reachable2,
                    num_shared_nodes);
            res = is_reachable(reachable1, state_nodes);
        } else {
            std::fill(reachable2.begin() + num_shared_nodes, reachable2.end(), true);
            updated_bw = bw_abstraction.propagate_reachability(
                    reachable1,
                    state_nodes,
                    reachable2,
                    num_shared_nodes);
            res = is_reachable(reachable1, bw_goal);
        }
        fw = !fw;
        reachable2.swap(reachable1);
    }
    return res;
}

}


#else


#include "utils.h"
#include "hyperabstraction_generator.h"
#include "generic_hyperabstraction_generator.h"
#include "../option_parser.h"
#include "../utils/timer.h"
#include "../utils/system.h"
#include <vector>
#include <string>
#include <algorithm>
#include <iostream>


namespace hyperabstraction {

int
get_max_distance(const std::vector<int>& distances,
                 const std::vector<Node>& nodes)
{
    int res = 0;
    for (int i = nodes.size() - 1; i >= 0; i--) {
        res = std::max(res, distances[nodes[i]]);
    }
    return res;
}

bool is_reachable(const std::vector<bool>& reachable,
                  const std::vector<Node>& nodes)
{
    for (int i = nodes.size() - 1; i >= 0; i--) {
        if (!reachable[nodes[i]]) {
            return false;
        }
    }
    return true;
}

void add_hyperabstraction_options_to_parser(options::OptionParser& parser)
{
    parser.add_option<std::shared_ptr<HyperabstractionGenerator> >("generator");
    std::vector<std::string> p{"false", "self", "inverse", "combination", "bidirectional"}; 
    parser.add_enum_option("prune_unreachable", p, "", "false");
}

std::shared_ptr<HyperabstractionGenerator>
get_hyperabstraction_generator_from_options(const options::Options& opts)
{
    return opts.get<std::shared_ptr<HyperabstractionGenerator> >("generator");
}

template<typename H, typename GF, typename GB>
std::pair<
    std::shared_ptr<HyperabstractionGeneratorResult<H> >,
    std::shared_ptr<HyperabstractionGeneratorResult<H> > >
get_hyperabstraction_from_options(
        const options::Options& opts,
        unsigned direction,
        unsigned shared_nodes,
        const GF& generate_forward,
        const GB& generate_backward)
{
    if (direction == 2
            && opts.get<int>("prune_unreachable") != 0
            && opts.get<int>("prune_unreachable") != 4) {
        std::cerr << "Bidirectional hyperabstractions are not compatible with prune_unreachable=";
        if (opts.get<int>("prune_unreachable") == 1) {
            std::cerr << "self" << std::endl;
        } else if (opts.get<int>("prune_unreachable") == 2) {
            std::cerr << "inverse" << std::endl;
        } else if (opts.get<int>("prune_unreachable") == 3) {
            std::cerr << "combination" << std::endl;
        }
        utils::exit_with(utils::ExitCode::CRITICAL_ERROR);
    }

    std::shared_ptr<HyperabstractionGeneratorResult<H> > fw = nullptr;
    std::shared_ptr<HyperabstractionGeneratorResult<H> > bw = nullptr;

    bool forward  = direction == 0 || direction == 2 || (opts.get<int>("prune_unreachable") > 1);
    bool backward = direction == 1 || direction == 2 || (opts.get<int>("prune_unreachable") > 1);
    unsigned num_nodes = 0;

    if (forward) {
        std::cout << "Constructing forward hyperabstraction..." << std::endl;
        fw = std::make_shared<HyperabstractionGeneratorResult<H> >(
                std::move(generate_forward()));
        fw->statistics.print(true);
        num_nodes = fw->abstraction->num_nodes();
    }

    if (backward) {
        std::cout << "Constructing backward hyperabstraction..." << std::endl;
        bw = std::make_shared<HyperabstractionGeneratorResult<H> >(
                std::move(generate_backward()));
        bw->statistics.print(false);
        num_nodes = std::max(num_nodes, bw->abstraction->num_nodes());
    }


    if (opts.get<int>("prune_unreachable")) {
        utils::Timer t;
        auto me = (direction == 0 ? fw : bw);
        std::vector<bool> reachable(num_nodes);
        if (opts.get<int>("prune_unreachable") > 1) {
            if (opts.get<int>("prune_unreachable") <= 3) {
                std::cout << "Unreachability pruning via the inverse direction..." << std::endl;
                auto inverse = (direction == 0 ? bw : fw);
                inverse->abstraction->compute_reachable(reachable, inverse->initial_nodes);
            } else {
                std::cout << "Unreachability pruning using bidirectional reachability computation..." << std::endl;
                fw->abstraction->compute_reachable(reachable, fw->initial_nodes);
                compute_bidirectional_reachability(
                        reachable,
                        fw->initial_nodes,
                        bw->goal_nodes,
                        fw->goal_nodes,
                        shared_nodes,
                        *(fw->abstraction),
                        *(bw->abstraction));
            }
            std::fill(reachable.begin() + shared_nodes, reachable.end(), true);
        } else {
            std::cout << "Unreachability pruning via distances in own abstraction..." << std::endl;
            me->abstraction->compute_reachable(reachable, me->initial_nodes);
        }
        if (direction == 2) {
            std::vector<bool> reachable2(num_nodes);

            fw->abstraction->propagate_reachability(reachable2, fw->initial_nodes, reachable, shared_nodes);
            unsigned transitions = fw->abstraction->num_transitions();
            unsigned unreachable = fw->abstraction->prune_unreachable(reachable2);
            std::cout << "Unreachable nodes (forward): " << unreachable
                      << " ("
                      << (((double) unreachable) / (double) fw->abstraction->num_nodes())
                      << ")"
                      << std::endl;
            std::cout << "Transitions removed (forward): "
                        << (transitions - fw->abstraction->num_transitions())
                        << " (" << (((double) (transitions - fw->abstraction->num_transitions())) / (double) transitions)
                        << ")" << std::endl;

            bw->abstraction->propagate_reachability(reachable2, bw->initial_nodes, reachable, shared_nodes);
            transitions = bw->abstraction->num_transitions();
            unreachable = bw->abstraction->prune_unreachable(reachable2);
            std::cout << "Unreachable nodes (backward): " << unreachable
                      << " ("
                      << (((double) unreachable) / (double) bw->abstraction->num_nodes())
                      << ")"
                      << std::endl;
            std::cout << "Transitions removed (backward): "
                        << (transitions - bw->abstraction->num_transitions())
                        << " (" << (((double) (transitions - bw->abstraction->num_transitions())) / (double) transitions)
                        << ")" << std::endl;
        } else {
            if (opts.get<int>("prune_unreachable") >= 3) {
                std::vector<bool> reachable2(me->abstraction->num_nodes());
                me->abstraction->propagate_reachability(
                        reachable2,
                        me->initial_nodes,
                        reachable,
                        shared_nodes);
                reachable.swap(reachable2);
            }
            unsigned transitions = me->abstraction->num_transitions();
            unsigned unreachable = me->abstraction->prune_unreachable(reachable);
            std::cout << "Unreachable nodes: " << unreachable
                      << " ("
                      << (((double) unreachable) / (double) me->abstraction->num_nodes())
                      << ")"
                      << std::endl;
            std::cout << "Transitions removed: "
                        << (transitions - me->abstraction->num_transitions())
                        << " (" << (((double) (transitions - me->abstraction->num_transitions())) / (double) transitions)
                        << ")" << std::endl;
        }
        std::cout << "Computed and pruned unreachable transitions in " << t << std::endl;
    }

    return std::pair<
        std::shared_ptr<HyperabstractionGeneratorResult<H> >,
        std::shared_ptr<HyperabstractionGeneratorResult<H> > >(
                fw,
                bw);
}

HyperabstractionGeneratorResult<Hyperabstraction>
get_hyperabstraction_from_options(
        const options::Options& opts,
        std::shared_ptr<HyperabstractionGenerator> generator,
        bool forward)
{
    auto get_forward = [&generator]() {
        return std::move(generator->generate_forward());
    };
    auto get_backward = [&generator]() {
        return std::move(generator->generate_backward());
    };
    auto t(get_hyperabstraction_from_options<Hyperabstraction>(opts, forward ? 0 : 1, generator->num_nodes(), get_forward, get_backward));
    if (forward) {
        return HyperabstractionGeneratorResult<Hyperabstraction>(std::move(*t.first));
    } else {
        return HyperabstractionGeneratorResult<Hyperabstraction>(std::move(*t.second));
    }
}

HyperabstractionGeneratorResult<LabeledHyperabstraction>
get_labeled_hyperabstraction_from_options(
        const options::Options& opts,
        std::shared_ptr<HyperabstractionGenerator> generator,
        bool forward)
{
    auto get_forward = [&generator]() {
        return std::move(generator->generate_forward_labeled());
    };
    auto get_backward = [&generator]() {
        return std::move(generator->generate_backward_labeled());
    };
    auto t(get_hyperabstraction_from_options<LabeledHyperabstraction>(opts, forward ? 0 : 1,  generator->num_nodes(), get_forward, get_backward));
    if (forward) {
        return HyperabstractionGeneratorResult<LabeledHyperabstraction>(std::move(*t.first));
    } else {
        return HyperabstractionGeneratorResult<LabeledHyperabstraction>(std::move(*t.second));
    }
}

std::pair<HyperabstractionGeneratorResult<Hyperabstraction>, HyperabstractionGeneratorResult<Hyperabstraction> >
get_bidirectional_hyperabstractions_from_options(
        const options::Options& opts,
        std::shared_ptr<HyperabstractionGenerator> generator)
{
    auto get_forward = [&generator]() {
        return std::move(generator->generate_forward());
    };
    auto get_backward = [&generator]() {
        return std::move(generator->generate_backward());
    };
    auto t(get_hyperabstraction_from_options<Hyperabstraction>(opts, 2, generator->num_nodes(), get_forward, get_backward));
    return std::make_pair
    (HyperabstractionGeneratorResult<Hyperabstraction>(std::move(*t.first)),
                          HyperabstractionGeneratorResult<Hyperabstraction>(std::move(*t.second)));
}

}

#endif
