#include "pattern_collection_generator_hyf.h"

#include "../abstract_task.h"
#include "../task_proxy.h"
#include "../causal_graph.h"
#include "../utils/hash.h"
#include "../utils/countdown_timer.h"

#include "../option_parser.h"
#include "../plugin.h"

#include <unordered_set>

namespace pdbs {

PatternCollectionGeneratorHYF::PatternCollectionGeneratorHYF(
        const options::Options& opts)
    : c_goal_pdb_max_size(opts.get<int>("goal_pdb_max_size"))
    , c_pdb_max_size(opts.get<int>("pdb_max_size"))
    , c_collection_max_size(opts.get<int>("collection_max_size"))
    , c_split(opts.get<double>("split"))
    , c_max_time(opts.get<double>("max_time"))
    , m_rng(opts.get<int>("seed"))
{}

PatternCollectionInformation
PatternCollectionGeneratorHYF::generate_collection(std::shared_ptr<AbstractTask> task)
{
    utils::CountdownTimer countdown(c_max_time);
    std::shared_ptr<PatternCollection> collection = std::make_shared<PatternCollection>();

    collection->reserve(task->get_num_goals());
    for (int i = task->get_num_goals() - 1; i >= 0; i--) {
        collection->emplace_back(Pattern{task->get_goal_fact(i).var});
    }
    m_rng.shuffle(*collection);

    std::vector<unsigned> domains(task->get_num_variables());
    for (int var = domains.size() - 1; var >= 0; var--) {
        domains[var] = task->get_variable_domain_size(var);
    }

    unsigned collection_size = 0;
    unsigned size = task->get_variable_domain_size(collection->back().back());
    int i = collection->size() - 2;
    for (; size < c_goal_pdb_max_size && i >= 0; i--) {
        int var = collection->at(i).back();
        unsigned d = domains[var];
        unsigned newsize = size * d;
        collection->at(i).swap(collection->at(i+1));
        if (newsize <= c_goal_pdb_max_size && collection_size + newsize < c_collection_max_size) {
            size = newsize;
            collection->at(i).push_back(var);
            if (i + 1 < (int) collection->size()) {
                collection->at(i+1).swap(collection->back());
            }
            collection->pop_back();
        } else {
            collection_size += d;
        }
    }
    if (i >= 0) {
        collection->at(0).swap(collection->at(i+1));
    }
    collection_size += size;

    const CausalGraph& cg = get_causal_graph(task.get());

    for (int i = collection->size() - 1; i >= 0; i--) {
           std::cout << "[";
           for (unsigned j = 0; j < collection->at(i).size(); j++) {
               std::cout << (j > 0 ? ", " : "") << collection->at(i)[j];
           }
           std::cout << "] -> " << (i > 0 ? domains[collection->at(i).back()] : size) << std::endl;
    }

    std::vector<bool> in_consideration(domains.size(), false);
    std::vector<int>  open(domains.size());
    std::unordered_set<int, utils::CollectionIdHasher<PatternCollection>, utils::CollectionIdComparator<PatternCollection>> unique(
            1012,
            utils::CollectionIdHasher<PatternCollection>(*collection),
            utils::CollectionIdComparator<PatternCollection>(*collection));
    for (int i = collection->size() - 1; i > 0; i--) { unique.insert(i); }
    while (collection_size < c_collection_max_size && !countdown.is_expired()) {
        collection->emplace_back();
        Pattern& goal_pattern = collection->at(0);
        m_rng.shuffle(goal_pattern);
        Pattern& pattern = collection->back();
        unsigned gpidx = std::max(1u, (unsigned) std::floor(c_split * goal_pattern.size()));
        pattern.insert(pattern.end(),
                       goal_pattern.begin(),
                       goal_pattern.begin() + gpidx);
        assert(!pattern.empty());
        std::fill(in_consideration.begin(), in_consideration.end(), false);
        open.clear();
        unsigned size = 1;
        const unsigned max_size = std::min(c_pdb_max_size, c_collection_max_size - collection_size);
        for (unsigned i = 0; i < pattern.size(); i++) {
            int var = pattern[i];
            unsigned newsize = size * domains[var];
            if (i > 0 && newsize > max_size) {
                pattern.resize(i);
                break;
            }
            size = newsize;
            in_consideration[var] = true;
        }
        for (unsigned i = 0; i < pattern.size(); i++) {
            const std::vector<int>& arcs = cg.get_predecessors(pattern[i]);
            for (const int& var : arcs) {
                if (!in_consideration[var]) {
                    in_consideration[var] = true;
                    open.push_back(var);
                }
            }
        }
        bool stop = true;
        while (size < max_size) {
            while (!open.empty() && size < max_size && !countdown.is_expired()) {
                int idx = m_rng(open.size());
                assert(idx < (int) open.size());
                int var = open[idx];
                if (idx + 1 < (int) open.size()) {
                    open[idx] = open.back();
                }
                open.pop_back();
                unsigned newsize = size * domains[var];
                if (newsize <= max_size) {
                    size = newsize;
                    pattern.push_back(var);
                    const std::vector<int>& arcs = cg.get_predecessors(var);
                    for (const int& var : arcs) {
                        assert(var < task->get_num_variables());
                        if (!in_consideration[var]) {
                            in_consideration[var] = true;
                            open.push_back(var);
                        }
                    }
                } else {
                    stop = false;
                }
            }
            while (gpidx < goal_pattern.size() && in_consideration[goal_pattern[gpidx]]) {
                gpidx++;
            }
            if (gpidx == goal_pattern.size()) {
                break;
            }
            in_consideration[goal_pattern[gpidx]] = true;
            open.push_back(goal_pattern[gpidx++]);
        }
        std::sort(pattern.begin(), pattern.end());
        if (!unique.insert(collection->size() - 1).second) {
            collection->pop_back();
        } else {
            collection_size += size;
            assert(!pattern.empty());
            std::cout << "[";
            for (unsigned i = 0; i < pattern.size(); i++) {
                std::cout << (i > 0 ? ", " : "") << pattern[i];
            }
            std::cout << "] -> " << size << std::endl;
        }
        if (stop) {
            break;
        }
    }
    std::sort(collection->front().begin(), collection->front().end());
    if (!unique.insert(0).second) {
        collection->front().swap(collection->back());
        collection->pop_back();
    }

    return PatternCollectionInformation(task, collection);
}

void
PatternCollectionGeneratorHYF::add_options_to_parser(options::OptionParser& parser)
{
    parser.add_option<int>("goal_pdb_max_size", "", "100000");
    parser.add_option<int>("pdb_max_size", "", "100000");
    parser.add_option<int>("collection_max_size", "", "1000000");
    parser.add_option<int>("seed", "", "1734");
    parser.add_option<double>("split", "", "0.3");
    parser.add_option<double>("max_time", "", "10");
}

}

static PluginShared<pdbs::PatternCollectionGenerator> _plugin("hyf",
        options::parse_shared<pdbs::PatternCollectionGenerator, pdbs::PatternCollectionGeneratorHYF>);
