#include "pattern_collection_generator_random.h"

#include "../abstract_task.h"
#include "../task_proxy.h"
#include "../causal_graph.h"
#include "../globals.h"
#include "../plugin.h"
#include "../utils/hash.h"
#include "../utils/ub_tree.h"
#include "../option_parser.h"

#include <cmath>
#include <cassert>
#include <algorithm>
#include <unordered_set>

namespace pdbs {

struct Hasher {
    const PatternCollection& patterns;
    Hasher(const PatternCollection& patterns) : patterns(patterns) {}
    unsigned operator()(int i) const { return utils::hash_sequence(patterns[i], patterns[i].size()); }
};

struct Comparator {
    const PatternCollection& patterns;
    Comparator(const PatternCollection& patterns) : patterns(patterns) {}
    bool operator()(int i, int j) const
    {
        const Pattern& p1 = patterns[i];
        const Pattern& p2 = patterns[j];
        return std::equal(p1.begin(), p1.end(), p2.begin(), p2.end());
    }
};

PatternCollectionGeneratorRandom::PatternCollectionGeneratorRandom(const options::Options& opts)
    : c_max_states_per_pattern(std::max(-1, opts.get<int>("pattern_max_states")))
    , c_max_states_overall(std::max(-1, opts.get<int>("max_states")))
    , c_p_strict(opts.get<bool>("strict"))
    , c_p_goal(opts.get<double>("goal"))
    , c_p_connectivity(opts.get<double>("connected"))
    , c_allow_subsets(opts.get<bool>("subsets"))
    , m_timeout(opts.get<double>("time_limit"))
    , m_rng(opts.get<int>("seed"))
{}

void
PatternCollectionGeneratorRandom::shuffle(std::vector<int>& set)
{
    for (int i = set.size() - 1; i > 0; i--) {
        int j = m_rng(i + 1);
        int old = set[i];
        set[i] = set[j];
        set[j] = old;
    }
}

int
PatternCollectionGeneratorRandom::sample(
        std::vector<int>& connected,
        std::vector<int>& goal,
        std::vector<int>& remaining,
        bool start)
{
    int res = -2;
    if (!start && m_rng() < c_p_connectivity) {
        res = -1;
        if (!connected.empty()) {
            res = connected.back();
            connected.pop_back();
        }
    }
    if (res == -2 || (res == -1 && !c_p_strict)) {
        if (m_rng() < c_p_goal) {
            res = -1;
            if (!goal.empty()) {
                res = goal.back();
                goal.pop_back();
            }
        }
    }
    if (res == -2 || (res == -1 && !c_p_strict)) {
        res = -1;
        if (!remaining.empty()) {
            res = remaining.back();
            remaining.pop_back();
        }
    }
    return res;
}

class PatternSelector {
public:
    virtual bool select(const Pattern& pattern) = 0;
};

class PatternSelectorUnique : public PatternSelector {
    std::unordered_set<int, Hasher, Comparator> pattern_set;
public:
    PatternSelectorUnique(const PatternCollection& collection)
        : pattern_set(128, Hasher(collection), Comparator(collection)) {}
    virtual bool select(const Pattern& ) override
    {
        auto it = pattern_set.insert(pattern_set.size());
        return it.second;
    }
};

class PatternSelectorSubset : public PatternSelector {
    utils::UBTreeMap<int, bool> pattern_set;
    virtual bool select(const Pattern& pattern) override
    {
        if (!pattern_set.contains_subset(pattern)
                && !pattern_set.contains_superset(pattern)) {
            pattern_set.insert(pattern, true);
            return true;
        }
        return false;
    }
};

PatternCollectionInformation
PatternCollectionGeneratorRandom::generate_collection(std::shared_ptr<AbstractTask> task)
{
    std::shared_ptr<PatternCollection> collection = std::make_shared<PatternCollection>();
    std::unique_ptr<PatternSelector> pattern_selector = nullptr;
    if (c_allow_subsets) {
        pattern_selector = std::unique_ptr<PatternSelector>(
                new PatternSelectorUnique(*collection));
    } else {
        pattern_selector = std::unique_ptr<PatternSelector>(
                new PatternSelectorSubset());
    }
    const CausalGraph cg = TaskProxy(*task).get_causal_graph();
    unsigned states_overall = 0;
    std::vector<int> v1;
    std::vector<int> v2;
    std::vector<int> v3;
    std::vector<bool> selected(task->get_num_variables(), false);
    std::vector<bool> connected(task->get_num_variables(), false);
    std::vector<bool> goal_variable(task->get_num_variables(), false);
    std::vector<int> allvars(task->get_num_variables());
    for (int i = task->get_num_goals() - 1; i >= 0; i--) {
        goal_variable[task->get_goal_fact(i).var] = true;
        allvars[i] = i;
    }
    m_timeout.start();
    while (!m_timeout.is_expired()) {
        v1.clear();
        v2.clear();
        v3.clear();
        std::fill(selected.begin(), selected.end(), false);
        std::fill(connected.begin(), connected.end(), false);
        for (int i = task->get_num_variables() - 1; i >= 0; i--) {
            if (goal_variable[i]) {
                v2.push_back(i);
            } else {
                v3.push_back(i);
            }
        }
        shuffle(v2);
        shuffle(v3);
        unsigned states = 1;
        collection->emplace_back();
        Pattern& pattern = collection->back();
        shuffle(allvars);
        while (true) {
            int var = sample(v1, v2, v3, states == 1);
            if (var == -1) {
                break;
            }
            assert(var >= 0);
            selected[var] = true;
            unsigned new_states = states * task->get_variable_domain_size(var);
            if (new_states <= c_max_states_per_pattern
                    && states_overall + new_states <= c_max_states_overall) {
                states = new_states;
                pattern.push_back(var);
                for (const auto& var2 : cg.get_predecessors(var)) {
                    connected[var2] = true;
                }
                for (const auto& var2 : cg.get_successors(var)) {
                    connected[var2] = true;
                }
                v1.clear();
                v2.clear();
                v3.clear();
                for (int i = task->get_num_variables() - 1; i >= 0; i--) {
                    int var2 = allvars[i];
                    if (selected[var2]) continue;
                    if (connected[var2]) v1.push_back(var2);
                    else if (goal_variable[var2]) v2.push_back(var2);
                    else v3.push_back(var2);
                }
            }
        }
        if (pattern.empty()) {
            collection->pop_back();
            break;
        } else {
            auto dump = [](const std::vector<int>& vars) {
                std::cout << "pattern = [ ";
                for (unsigned i = 0; i < vars.size(); i++) {
                    std::cout << vars[i] << " ";
                } 
                std::cout << "]" << std::endl;
            };
            std::sort(pattern.begin(), pattern.end());
            if (!pattern_selector->select(pattern)) {
                collection->pop_back();
            } else {
                states_overall += states;
                dump(pattern);
            }
        }
    }
    std::cout << "Generated " << collection->size() << " patterns with states = " << states_overall << std::endl;
    return PatternCollectionInformation(task, collection);
}

void
PatternCollectionGeneratorRandom::add_options_to_parser(options::OptionParser& parser)
{
    parser.add_option<int>    ("pattern_max_states", "", "100000");
    parser.add_option<int>    ("max_states", "", "5000000");
    parser.add_option<bool>   ("strict", "", "false");
    parser.add_option<bool>   ("subsets", "", "false");
    parser.add_option<double> ("goal", "", "1");
    parser.add_option<double> ("connected", "", "0.85");
    parser.add_option<double> ("time_limit", "", "2");
    parser.add_option<int>    ("seed", "", "1734");
}

}

static PluginShared<pdbs::PatternCollectionGenerator> _plugin("random", options::parse_shared<pdbs::PatternCollectionGenerator, pdbs::PatternCollectionGeneratorRandom>);
