#ifndef UTILS_UB_TREE_H
#define UTILS_UB_TREE_H

#include <vector>
#include <utility>
#include <deque>
#include <functional>
#include <iostream>

#include "../segmented_vector.h"

namespace utils
{

template<typename K, typename V, typename Less = std::less<K>, typename EqualTo = std::equal_to<K> >
class UBTreeMap
{
private:
    static Less less;
    static EqualTo equal_to;

    struct Node;
    typedef std::vector<std::pair<K, Node*> > NodeList;
    struct Node {
        V* m;
        NodeList s;
        Node() : m(NULL) {}
    };
public:
    class const_iterator {
    friend class UBTreeMap;
    private:
        struct pos_t {
            const K* key;
            const Node* node;
            unsigned i;
            pos_t(const Node *node) : key(NULL), node(node), i(0) {}
            pos_t(const K* key, const Node *node) : key(key), node(node), i(0) {}
            bool operator==(const pos_t &other) const { return key == other.key && node == other.node && i == other.i; }
        };
        std::deque<pos_t> q;
        explicit const_iterator();
        explicit const_iterator(const Node* root);
    public:
        const_iterator(const const_iterator& other);
        const_iterator(const const_iterator&& other);
        const_iterator& operator++(int);
        const_iterator& skip();
        bool operator==(const const_iterator& other) const;
        bool operator!=(const const_iterator& other) const;
        bool is_set_completed() const;
        bool is_root() const;
        bool is_terminal() const;
        const K& key() const;
        const V& value() const;
    };
private:
    static Node* get(Node* node, const K& key);
    static Node* get(Node* node, const std::vector<K>& keys);
    static const Node* get(const Node* node, const K& key);
    static const Node* get(const Node* node, const std::vector<K>& keys);
    static const V* get_any_subset(const Node* node, const std::vector<K>& keys, unsigned i);
    static bool contains_superset(const Node* node, const std::vector<K>& keys, unsigned i);
    template<typename F>
    static void for_every_subset_do(const Node* node, const std::vector<K>& keys, unsigned i, const F& fun);
    template<typename F>
    static void for_every_superset_do(const Node* node, const std::vector<K>& keys, unsigned i, const F& fun);
    template<typename F>
    static int remove(Node* node, const std::vector<K> &keys, unsigned i, const F& fun);

    Node* get_or_create(Node* node, const K& key);


    SegmentedVector<Node> m_nodes;
    SegmentedVector<V> m_leaf_data;
    size_t m_num_nodes;
    size_t m_num_leafs;

    Node m_root;
public:
    UBTreeMap();
    UBTreeMap(UBTreeMap &&other);
    std::pair<V*, bool> insert(const std::vector<K>& key, const V& v);
    bool remove(const std::vector<K>& key);
    template<typename F>
    bool remove(const std::vector<K>& key, const F& fun);
    const V* get(const std::vector<K> &key) const;
    bool contains(const std::vector<K>& key) const;
    bool contains_subset(const std::vector<K>& key) const;
    bool contains_superset(const std::vector<K> &key) const;
    template<typename F>
    void for_every_subset_do(const std::vector<K>& key,
                             const F& fun) const;
    void get_values_for_all_subsets(const std::vector<K>& key,
                                    std::vector<V>& result) const;
    template<typename F>
    void for_every_superset_do(const std::vector<K>& key,
                               const F& fun) const;
    
    void clear();
    size_t size() const;

    const_iterator begin() const;
    const_iterator end() const;

    void graphviz(std::ostream &out = std::cout) const;
};

}

#include "ub_tree.cc"

#endif
